/*
 * This program sets up the following network QoS elements.
 * +- Root HTB QDisc
 *    +- An HTB class that should have rate and ceiling set to the interface speed.
 *       +- A priomap QDisc with the default mappings and 3 bands.
 *          +- A SFQ QDisc for the first band.
 *          +- A SFQ QDisc for the second band.
 *          +- A SFQ QDisc for the third band.
 *
 * This is the arrangment I use at my network's edge. Most applications set the IP TOS
 * bits properly (most importantly OpenSSH does). The priomap looks at this field and
 * prioritizes the traffic. This layout seems to keep SSH latency very low while the
 * pipe is full.
 */
#include <glib.h>

#include "lql-con.h"
#include "lql-qdisc.h"
#include "lql-qdisc-htb.h"
#include "lql-qdisc-sfq.h"
#include "lql-qdisc-priomap.h"
#include "lql-class.h"
#include "lql-class-htb.h"
#include "lql-interface.h"

int
main(int argc,
     char *argv[])
{
	LQLCon *con = NULL;
	LQLQDiscHTB *qdisc = NULL;
	LQLQDiscSFQ *sfq1 = NULL, *sfq2 = NULL, *sfq3 = NULL;
	LQLQDiscPriomap *prio = NULL;
	LQLClassHTB *htb = NULL;
	LQLInterface *interface = NULL;
	GList *list = NULL;

	g_type_init();

	con = (LQLCon *)lql_con_new();

	// Get the interfaces and find the index of eth0.
	list = lql_con_get_interfaces(con);
	while (list != NULL) {
		interface = list->data;
		if (g_strcasecmp("eth0", lql_interface_get_name(interface)) == 0) {
			break;
		} else {
			g_object_unref(interface);
		}

		list = list->next;
	}
	g_list_free(list);

	// Add the HTB QDisc.
	qdisc = lql_qdisc_htb_new();
	lql_qdisc_set_parent(LQL_QDISC(qdisc), NULL);
	lql_qdisc_set_id(LQL_QDISC(qdisc), 1);
	lql_qdisc_htb_set_defaultClass(LQL_QDISC_HTB(qdisc), 1);
	lql_element_set_interface(LQL_ELEMENT(qdisc), interface);
	if (!lql_qdisc_add(LQL_QDISC(qdisc), con)) {
		g_printerr("Could not add HTB QDisc.\n");
		return(-1);
	}

	// Add the HTB Class.
	htb = lql_class_htb_new();
	lql_class_set_parent(LQL_CLASS(htb), LQL_ELEMENT(qdisc));
	lql_class_set_priority(LQL_CLASS(htb), 1);
	lql_class_set_id(LQL_CLASS(htb), 1);
	lql_element_set_interface(LQL_ELEMENT(htb), LQL_INTERFACE(interface));
	lql_class_htb_set_rate(htb, 1250000);
	lql_class_htb_set_ceiling(htb, 1250000);
	if (!lql_class_add(LQL_CLASS(htb), con)) {
		g_printerr("Could not add class.\n");
		return(-1);
	}

	// Add the Priomap QDisc.
	prio = lql_qdisc_priomap_new();
	lql_qdisc_set_parent(LQL_QDISC(prio), LQL_ELEMENT(htb));
	lql_qdisc_set_id(LQL_QDISC(prio), 2);
	lql_element_set_interface(LQL_ELEMENT(prio), LQL_INTERFACE(interface));
	lql_qdisc_priomap_set_bands(LQL_QDISC_PRIOMAP(prio), 3);
	if (!lql_qdisc_add(LQL_QDISC(prio), con)) {
		g_printerr("Could not add priomap qdisc.\n");
		return(-1);
	}

	sfq1 = lql_qdisc_sfq_new();
	lql_qdisc_set_parent(LQL_QDISC(sfq1), LQL_ELEMENT(prio));
	lql_qdisc_set_parent_class_id(LQL_QDISC(sfq1), 1);
	lql_qdisc_set_id(LQL_QDISC(sfq1), 5);
	lql_element_set_interface(LQL_ELEMENT(sfq1), LQL_INTERFACE(interface));
	lql_qdisc_sfq_set_perturb(LQL_QDISC_SFQ(sfq1), 10);
	if (!lql_qdisc_add(LQL_QDISC(sfq1), con)) {
		g_printerr("Could not add sfq1.\n");
		return(-1);
	}

	sfq2 = lql_qdisc_sfq_new();
	lql_qdisc_set_parent(LQL_QDISC(sfq2), LQL_ELEMENT(prio));
	lql_qdisc_set_parent_class_id(LQL_QDISC(sfq2), 2);
	lql_qdisc_set_id(LQL_QDISC(sfq2), 6);
	lql_element_set_interface(LQL_ELEMENT(sfq2), LQL_INTERFACE(interface));
	lql_qdisc_sfq_set_perturb(LQL_QDISC_SFQ(sfq2), 10);
	if (!lql_qdisc_add(LQL_QDISC(sfq2), con)) {
		g_printerr("Could not add sfq2.\n");
		return(-1);
	}

	sfq3 = lql_qdisc_sfq_new();
	lql_qdisc_set_parent(LQL_QDISC(sfq3), LQL_ELEMENT(prio));
	lql_qdisc_set_parent_class_id(LQL_QDISC(sfq3), 3);
	lql_qdisc_set_id(LQL_QDISC(sfq3), 7);
	lql_element_set_interface(LQL_ELEMENT(sfq3), LQL_INTERFACE(interface));
	lql_qdisc_sfq_set_perturb(LQL_QDISC_SFQ(sfq3), 10);
	if (!lql_qdisc_add(LQL_QDISC(sfq3), con)) {
		g_printerr("Could not add sfq3.\n");
		return(-1);
	}
	
	// Cleanup.
	g_object_unref(con);
	g_object_unref(interface);
	g_object_unref(qdisc);
	g_object_unref(htb);
	g_object_unref(prio);
	g_object_unref(sfq1);
	g_object_unref(sfq2);
	g_object_unref(sfq3);

	return(0);
}
