/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-classifier-fwmark.h"
#include "lql-classifier-fwmark-private.h"
#include "lql_classifier_fwmark_helper.h"

/********************************************************************
 * Non-Virutal API
 ********************************************************************/

/**
 * lql_classifier_fwmark_new:
 *
 * Build a new instance of LQLClassifierFwmark. 
 * 
 * Return value: A new LQLClassifierFwmark instance.
 **/
LQLClassifierFwmark*
lql_classifier_fwmark_new(void)
{
        return(LQL_CLASSIFIER_FWMARK(g_object_new(LQL_TYPE_CLASSIFIER_FWMARK, NULL)));
}

/**
 * lql_classifier_fwmark_get_mark:
 * @self: The LQLClassifierFwmark object to work with.
 * 
 * Get the mark.
 * 
 * Return value: The mark for this filter.
 **/
guint32
lql_classifier_fwmark_get_mark(LQLClassifierFwmark *self)
{
        guint32 mark = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_CLASSIFIER(self), "id", &mark, NULL);

        return(mark);
}

/**
 * lql_classifier_fwmark_set_mark:
 * @self: The LQLClassifier object to work with.
 * @mark: The mark value.
 * 
 * Set the mark value for this classifier.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_classifier_fwmark_set_mark(LQLClassifierFwmark *self,
		       guint32 mark)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASSIFIER(self), "id", mark, NULL);
        
        return(TRUE);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/**
 * lql_classifier_fwmark_fill_from_nlhdr:
 * @self: The LQLClassifier object to work with.
 * @hdr: The nl packet to get the queue information from.
 *
 * This method takes a nlmsg from the kernel and set this
 * objects instance variables to match it. This is used
 * to create objects that match the ones currently active
 * in the system. This overrides the default LQLClassifier
 * implementation of this method.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_fwmark_fill_from_nlhdr(LQLClassifier *self,
				      const struct nlmsghdr *hdr)
{
	g_assert(self != NULL);
	g_assert(hdr != NULL);

	return(lql_classifier_fwmark_helper_fill_from_nlhdr(LQL_CLASSIFIER_FWMARK(self), hdr));
}

/**
 * lql_classifier_fwmark_add:
 * @self: The LQLClassifier object to add to the system.
 * @con: The LQLCon connection to use when adding this queueing discipline.
 *
 * Add this HTB queueing discipline to the system. This overrides
 * the default implementation of this function in LQLClassifier.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_fwmark_add(LQLClassifier *self,
			  LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_classifier_fwmark_helper_add(LQL_CLASSIFIER_FWMARK(self), con));
}

/**
 * lql_classifier_fwmark_modify:
 * @self: The LQLClassifier object to modify.
 * @con: The LQLCon connection to use when adding this classifier.
 *
 * Modify the HTB queueing discipline in the system to match the current
 * object state. This method overrides the default
 * implementation in LQLClassifier.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_fwmark_modify(LQLClassifier *self,
			     LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_classifier_fwmark_helper_modify(LQL_CLASSIFIER_FWMARK(self), con));
}

/********************************************************************
 * Object stuff.
 ********************************************************************/
static void
lql_classifier_fwmark_instance_init(GTypeInstance *instance,
				    gpointer g_class)
{
	LQLClassifierFwmark *self = (LQLClassifierFwmark *)instance;
	self->private = g_new0(LQLClassifierFwmarkPrivate, 1);
	self->private->dispose_has_run = FALSE;
}

static void
lql_classifier_fwmark_set_property(GObject *object,
				   guint property_id,
				   const GValue *value,
				   GParamSpec *pspec)
{
        //LQLClassifierFwmark *self = (LQLClassifierFwmark *) object;

        switch (property_id) {
        default:
                g_assert(FALSE);
                break;
        }
}

static void
lql_classifier_fwmark_get_property(GObject *object,
				   guint property_id,
				   GValue *value,
				   GParamSpec *pspec)
{
        //LQLClassifierFwmark *self = (LQLClassifierFwmark *) object;

        switch (property_id) {
        default:
                g_assert(FALSE);
                break;
        }
}

static GObject *
lql_classifier_fwmark_constructor(GType type,
				  guint n_construct_properties,
				  GObjectConstructParam *construct_properties)
{
        GObject *obj = NULL;

        {
                /* Invoke parent constructor. */
		LQLClassifierFwmarkClass *klass;
		GObjectClass *parent_class;  
		klass = LQL_CLASSIFIER_FWMARK_CLASS(g_type_class_peek(LQL_TYPE_CLASSIFIER_FWMARK));
		parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
		obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_classifier_fwmark_dispose(GObject *obj)
{
        LQLClassifierFwmark *self = (LQLClassifierFwmark *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_classifier_fwmark_finalize(GObject *obj)
{
        LQLClassifierFwmark *self = (LQLClassifierFwmark *)obj;

        /*
         * Here, complete object destruction.
         */
        
        g_free (self->private);
}


static void
lql_classifier_fwmark_class_init(gpointer g_class,
				 gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        LQLClassifierClass *klass = LQL_CLASSIFIER_CLASS(g_class);

        gobject_class->set_property = lql_classifier_fwmark_set_property;
        gobject_class->get_property = lql_classifier_fwmark_get_property;
        gobject_class->dispose = lql_classifier_fwmark_dispose;
        gobject_class->finalize = lql_classifier_fwmark_finalize;
        gobject_class->constructor = lql_classifier_fwmark_constructor;

	/* Set the virtual methods. */
	klass->fill_from_nlhdr = lql_classifier_fwmark_fill_from_nlhdr;
	klass->add = lql_classifier_fwmark_add;
	klass->modify = lql_classifier_fwmark_modify;

	/* Setup the properties. */
}

GType
lql_classifier_fwmark_get_type(void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLClassifierFwmarkClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_classifier_fwmark_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLClassifierFwmark),
                        0,      /* n_preallocs */
                        lql_classifier_fwmark_instance_init    /* instance_init */
                };
                type = g_type_register_static (LQL_TYPE_CLASSIFIER,
                                               "LQLClassifierFwmark",
                                               &info, 0);
        }
        return type;
}
