/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-classifier.h"
#include "lql-classifier-private.h"
#include "lql_classifier_helper.h"

/********************************************************************
 * Non-Virtual API
 ********************************************************************/

/**
 * lql_classifier_new:
 *
 * Build a new instance of LQLClassifier. 
 * 
 * Return value: A new LQLClassifier instance.
 **/
LQLClassifier*
lql_classifier_new(void)
{
        return(LQL_CLASSIFIER(g_object_new(LQL_TYPE_CLASSIFIER, NULL)));
}

/**
 * lql_classifier_set_interface:
 * @self: The LQLClassifier instance.
 * @interface: The LQLInterface object.
 * 
 * Set the network interface that this classifier is attached to.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_set_interface(LQLClassifier *self,
			     LQLInterface *interface)
{
	g_assert(self != NULL);
	g_assert(interface != NULL);

	if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASSIFIER(self), "ifIndex",
		     lql_interface_get_ifIndex(LQL_INTERFACE(interface)), NULL);
        
        return(TRUE);
}

/**
 * lql_classifier_get_ifIndex:
 * @self: The LQLClassifier object to work with.
 * 
 * Get the interface index this Classifier is attached to.
 * 
 * Return value: The classifier index, -1 on error.
 **/
int
lql_classifier_get_ifIndex(LQLClassifier *self)
{
        int ifIndex = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }

        g_object_get(LQL_CLASSIFIER(self), "ifIndex", &ifIndex, NULL);

        return(ifIndex);
}

/**
 * lql_classifier_set_ifIndex:
 * @self: The LQLClassifier object to work with.
 * @ifIndex: The index of a network interface.
 * 
 * Set the network interface index that this Classifier is attached to.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_classifier_set_ifIndex(LQLClassifier *self,
		      int ifIndex)
{
        g_assert(self != NULL);
        g_assert(ifIndex > 0);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASSIFIER(self), "ifIndex", ifIndex, NULL);
        
        return(TRUE);
}

/**
 * lql_classifier_get_parentHandle:
 * @self: The LQLClassifier object to work with.
 * 
 * Get the handle of the parent of this Classifier.
 * 
 * Return value: The classifier parent, -1 on error.
 **/
guint32
lql_classifier_get_parentHandle(LQLClassifier *self)
{
        guint32 parent = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_CLASSIFIER(self), "parent", &parent, NULL);

        return(parent);
}

/**
 * lql_classifier_set_parentHandle:
 * @self: The LQLClassifier instance.
 * @parent: The handle of the parent.
 * 
 * Set the handle of the parent of this LQLClassifier. Usually you will want to use
 * lql_classifier_set_parent() instead of this function.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_set_parentHandle(LQLClassifier *self,
				guint32 parent)
{
	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }

        g_object_set(LQL_CLASSIFIER(self), "parent", parent, NULL);

	return(TRUE);
}

/**
 * lql_classifier_set_parent:
 * @self: The LQLClassifier object to work with.
 * @parent: The LQLElement that is the parent.
 * 
 * Set the parent field of this object.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_classifier_set_parent(LQLClassifier *self,
			  LQLElement *parent)
{
        g_assert(self != NULL);
	g_assert(parent != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASSIFIER(self), "parent",
		     lql_element_get_handle(LQL_ELEMENT(parent)), NULL);
        
        return(TRUE);
}

/**
 * lql_classifier_get_id:
 * @self: The LQLClassifier object to work with.
 * 
 * Get the ID of this Classifier.
 * 
 * Return value: The classifier ID.
 **/
guint32
lql_classifier_get_id(LQLClassifier *self)
{
        guint32 id = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_CLASSIFIER(self), "id", &id, NULL);

        return(id);
}

/**
 * lql_classifier_set_id:
 * @self: The LQLClassifier object to work with.
 * @id: The value to set the ID instance variable to.
 * 
 * Set the ID field of this object.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_classifier_set_id(LQLClassifier *self,
		      guint32 id)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASSIFIER(self), "id", id, NULL);
        
        return(TRUE);
}

/**
 * lql_classifier_get_classHandle:
 * @self: The LQLClassifier object to work with.
 * 
 * Get the handle of the class that this LQLClassifer will direct packets to.
 * 
 * Return value: The classifier handle.
 **/
guint32
lql_classifier_get_classHandle(LQLClassifier *self)
{
        guint32 class = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_CLASSIFIER(self), "class", &class, NULL);

        return(class);
}

/**
 * lql_classifier_set_classHandle:
 * @self: The LQLClassifier instance.
 * @class: The handle of the class.
 * 
 * Set the handle of the class that this classifier will direct packets to.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_set_classHandle(LQLClassifier *self,
			       guint32 class)
{
	g_assert(self != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASSIFIER(self), "class", class, NULL);
        
        return(TRUE);
}

/**
 * lql_classifier_set_class:
 * @self: The LQLClassifier object to work with.
 * @class: The LQLClass to set.
 * 
 * Set the CLASS field of this object.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_classifier_set_class(LQLClassifier *self,
			 LQLClass *class)
{
        g_assert(self != NULL);
	g_assert(class != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASSIFIER(self), "class",
		     lql_element_get_handle(LQL_ELEMENT(class)), NULL);
        
        return(TRUE);
}

/**
 * lql_classifier_get_priority:
 * @self: The LQLClassifier object to work with.
 * 
 * Get the ID of the priority of this Classifier.
 * 
 * Return value: The classifier priority, -1 on error.
 **/
unsigned int
lql_classifier_get_priority(LQLClassifier *self)
{
        unsigned int priority = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_CLASSIFIER(self), "priority", &priority, NULL);

        return(priority);
}

/**
 * lql_classifier_set_priority:
 * @self: The LQLClassifier object to work with.
 * @priority: The index to set the priority instance variable to.
 * 
 * Set the priority field of this object.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_classifier_set_priority(LQLClassifier *self,
		     unsigned int priority)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASSIFIER(self), "priority", priority, NULL);
        
        return(TRUE);
}

/**
 * lql_classifier_get_protocol:
 * @self: The LQLClassifier object to work with.
 * 
 * Get the protocol number that this classifier is working with.
 * 
 * Return value: The classifier protocol, -1 on error.
 **/
unsigned int
lql_classifier_get_protocol(LQLClassifier *self)
{
        unsigned int protocol = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_CLASSIFIER(self), "protocol", &protocol, NULL);

        return(protocol);
}

/**
 * lql_classifier_set_protocol:
 * @self: The LQLClassifier object to work with.
 * @protocol: The index to set the protocol instance variable to.
 * 
 * Set the protocol field of this object.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_classifier_set_protocol(LQLClassifier *self,
		     unsigned int protocol)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_CLASSIFIER(self), "protocol", protocol, NULL);
        
        return(TRUE);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/**
 * lql_classifier_fill_from_nlhdr:
 * @self: The LQLClassifier object to work with.
 * @hdr: The classifier information from the kernel.
 * 
 * Fill in the instance variables for this instance of LQLClassifier
 * to match the network classifier information in hdr.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_fill_from_nlhdr(LQLClassifier *self,
			  const struct nlmsghdr *hdr)
{
        g_assert(self != NULL);
        g_assert(hdr != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_CLASSIFIER_GET_CLASS(self)->fill_from_nlhdr(self, hdr));
}

/**
 * lql_classifier_add:
 * @self: The LQLClassifier object to add to the system.
 * @con: The LQLCon connection to use when adding this queueing discipline.
 *
 * Add this queueing discipline to the system.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_add(LQLClassifier *self,
	      LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_CLASSIFIER_GET_CLASS(self)->add(self, con));
}

/**
 * lql_classifier_modify:
 * @self: The LQLClassifier object to modify.
 * @con: The LQLCon connection to use when adding this classifier.
 *
 * Modify the queueing discipline in the system to match the current
 * object state.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_modify(LQLClassifier *self,
		 LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_CLASSIFIER_GET_CLASS(self)->modify(self, con));
}

/**
 * lql_classifier_delete:
 * @self: The LQLClassifier object to delete from the system.
 * @con: The LQLCon connection to use when deleting this queueing discipline.
 *
 * Delete this queueing discipline from the system.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_classifier_delete(LQLClassifier *self,
		 LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_CLASSIFIER_GET_CLASS(self)->delete(self, con));
}

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_CLASSIFIER_IFINDEX = 1,
	LQL_CLASSIFIER_ID = 2,
	LQL_CLASSIFIER_PARENT = 3,
	LQL_CLASSIFIER_CLASS = 4,
	LQL_CLASSIFIER_PRIORITY = 5,
	LQL_CLASSIFIER_PROTOCOL = 6,
};

static void
lql_classifier_instance_init(GTypeInstance *instance,
			gpointer g_class)
{
        LQLClassifier *self = (LQLClassifier *)instance;
        self->private = g_new0 (LQLClassifierPrivate, 1);
        self->private->dispose_has_run = FALSE;

	self->private->protocol = 8;
}

static void
lql_classifier_set_property(GObject *object,
		       guint property_id,
		       const GValue *value,
		       GParamSpec *pspec)
{
        LQLClassifier *self = (LQLClassifier *) object;

        switch (property_id) {
        case LQL_CLASSIFIER_IFINDEX: 
        {
                self->private->ifIndex = g_value_get_int(value);
                break;
        }
	case LQL_CLASSIFIER_ID:
	{
		self->private->id = g_value_get_uint(value);
		break;
	}
	case LQL_CLASSIFIER_PARENT:
	{
		self->private->parent = g_value_get_uint(value);
		break;
	}
	case LQL_CLASSIFIER_CLASS:
	{
		self->private->class = g_value_get_uint(value);
		break;
	}
	case LQL_CLASSIFIER_PRIORITY:
	{
		self->private->priority = g_value_get_uint(value);
		break;
	}
	case LQL_CLASSIFIER_PROTOCOL:
	{
		self->private->protocol = g_value_get_uint(value);
		break;
	}
        default:
                g_assert (FALSE);
                break;
        }
}

static void
lql_classifier_get_property(GObject *object,
		       guint property_id,
		       GValue *value,
		       GParamSpec *pspec)
{
        LQLClassifier *self = (LQLClassifier *) object;

        switch (property_id) {
        case LQL_CLASSIFIER_IFINDEX: 
        {
                g_value_set_int(value, self->private->ifIndex);
                break;
        }
	case LQL_CLASSIFIER_ID:
	{
		g_value_set_uint(value, self->private->id);
		break;
	}
	case LQL_CLASSIFIER_PARENT:
	{
		g_value_set_uint(value, self->private->parent);
		break;
	}
	case LQL_CLASSIFIER_CLASS:
	{
		g_value_set_uint(value, self->private->class);
		break;
	}
	case LQL_CLASSIFIER_PRIORITY:
	{
		g_value_set_uint(value, self->private->priority);
		break;
	}
	case LQL_CLASSIFIER_PROTOCOL:
	{
		g_value_set_uint(value, self->private->protocol);
		break;
	}
        default:
                g_assert (FALSE);
                break;
        }
}

static GObject *
lql_classifier_constructor(GType type,
		      guint n_construct_properties,
		      GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                LQLClassifierClass *klass;
                GObjectClass *parent_class;  
                klass = LQL_CLASSIFIER_CLASS(g_type_class_peek(LQL_TYPE_CLASSIFIER));
                parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
                obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_classifier_dispose (GObject *obj)
{
        LQLClassifier *self = (LQLClassifier *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_classifier_finalize (GObject *obj)
{
        LQLClassifier *self = (LQLClassifier *)obj;

        /*
         * Here, complete object destruction.
         * You might not need to do much...
         */

        
        g_free (self->private);
}

static void
lql_classifier_class_init (gpointer g_class,
		      gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        LQLClassifierClass *klass = LQL_CLASSIFIER_CLASS(g_class);
        GParamSpec *pspec = NULL;

        gobject_class->set_property = lql_classifier_set_property;
        gobject_class->get_property = lql_classifier_get_property;
        gobject_class->dispose = lql_classifier_dispose;
        gobject_class->finalize = lql_classifier_finalize;
        gobject_class->constructor = lql_classifier_constructor;

	/* Set the virtual methods. */
	klass->fill_from_nlhdr = NULL;
	klass->add = NULL;
	klass->modify = NULL;
	klass->delete = lql_classifier_helper_delete;

	/* Setup the properties. */
        pspec = g_param_spec_int("ifIndex",
                                 "Network interface index",
                                 "Set/Get the network interface this LQLClassifier is "
				 "attached to",
                                 G_MININT /* minimum value */,
                                 G_MAXINT /* maximum value */,
                                 0 /* default value */,
                                 G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASSIFIER_IFINDEX,
                                        pspec);

        pspec = g_param_spec_uint("id",
				  "ID for this Classifier",
				  "Set/Get the ID for this Classifier",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASSIFIER_ID,
                                        pspec);

        pspec = g_param_spec_uint("parent",
				  "ID for the parent of this classifier",
				  "Set/Get the parent ID for this classifier",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASSIFIER_PARENT,
                                        pspec);

        pspec = g_param_spec_uint("class",
				  "The class to direct matching packets to",
				  "Set/Get the ID of the class to direct matching"
				  " packets to",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASSIFIER_CLASS,
                                        pspec);

        pspec = g_param_spec_uint("priority",
				  "Priority of this classifier",
				  "Set/Get the priority of this classifier",
				  0 /* minimum value */,
				  G_MAXUINT16 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASSIFIER_PRIORITY,
                                        pspec);

        pspec = g_param_spec_uint("protocol",
				  "Protocol ID for this classifier",
				  "Set/Get the protocol ID for this classifier",
				  0 /* minimum value */,
				  G_MAXUINT16 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_CLASSIFIER_PROTOCOL,
                                        pspec);
}

GType lql_classifier_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLClassifierClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_classifier_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLClassifier),
                        0,      /* n_preallocs */
                        lql_classifier_instance_init    /* instance_init */
                };
                type = g_type_register_static(G_TYPE_OBJECT,
					      "LQLClassifier",
					      &info, 0);
        }

        return type;
}
