/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "nl.h"

#include "lql-con.h"
#include "lql-con-private.h"
#include "lql_con_helper.h"

/********************************************************************
 * Non-Virutal API
 ********************************************************************/

/**
 * lql_con_new:
 *
 * Build a new instance of LQLCon. 
 * 
 * Return value: A new LQLCon instance.
 **/
LQLCon*
lql_con_new(void)
{
        return(LQL_CON(g_object_new(LQL_TYPE_CON, NULL)));
}

/**
 * lql_con_get_inc_sequence:
 * @self: The LQLCon object to work with.
 * 
 * Get the next sequence to use when communicating with the kernel.
 * 
 * Return value: The next sequence number, -1 on error.
 **/
int
lql_con_get_inc_sequence(LQLCon *self)
{
        int seq = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }

        self->private->seq++;
        g_object_get(LQL_CON(self), "sequence", &seq, NULL);

        return(seq);
}

/**
 * lql_con_get_tickInUsec:
 * @self: The LQLCon instance.
 * 
 * Fetch the number of ticks in a usec on this system.
 * 
 * Return value: The number of ticks in a usec, -1 on error.
 **/
double
lql_con_get_tickInUsec(LQLCon *self)
{
	double tickInUsec = 0;
	
	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }

	g_object_get(LQL_CON(self), "tick-in-usec", &tickInUsec, NULL);
	
	return(tickInUsec);
}

/**
 * lql_con_get_hz:
 * @self: The LQLCon instance.
 * 
 * Get the HZ for this system.
 * 
 * Return value: The HZ, -1 on error.
 **/
double
lql_con_get_hz(LQLCon *self)
{
	double hz = 0;
	
	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }

	g_object_get(LQL_CON(self), "hz", &hz, NULL);
	
	return(hz);
}

/**
 * lql_con_get_nlcon:
 * @self: The LQLCon instance.
 * 
 * Fetch a pointer to the NLCon that this LQLCon is using to communicate with the kernel.
 * 
 * Return value: A pointer to a NLCon* instance, -1 on error.
 **/
NLCon*
lql_con_get_nlcon(LQLCon *self)
{
	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(NULL);
        }

	return(self->private->con);
}

/**
 * lql_con_get_interface:
 * @self: The LQLCon to query.
 * @index: The index of the interface you want.
 *
 * This function returns a LQLInterface object with
 * represents the interface at the specified index.
 *
 * Returns: A LQLInterface object.
 **/
LQLInterface*
lql_con_get_interface(LQLCon *self,
		      int index)
{
	g_assert(self != NULL);
	g_assert(index > 0);

	return(lql_con_helper_get_interface(self, index));
}

/**
 * lql_con_get_interface_by_name:
 * @self: LQLCon instance.
 * @name: The name to search for.
 * 
 * Find the network interface by name.
 * 
 * Return value: LQLInterface instance or NULL on error.
 **/
LQLInterface*
lql_con_get_interface_by_name(LQLCon *self,
			      const gchar *name)
{
	GList *list = NULL, *list1 = NULL;
	LQLInterface *interface = NULL, *ret = NULL;

	g_assert(self != NULL);
	g_assert(name != NULL);
	
	list = lql_con_get_interfaces(self);
	list1 = list;
	while (list != NULL) {
		interface = list->data;
		if (g_strcasecmp(name, lql_interface_get_name(interface)) == 0) {
			ret = interface;
			break;
		} else {
			g_object_unref(interface);
		}

		list = list->next;
	}

	/* Now go over the rest of the list to unref the objects so that we don't leak
	   them. This loop would not be necessary if the break was not in the above loop.
	   I think this second loop is probably faster than g_strcasecmp() being called
	   a bunch of times.*/
	list = list->next;
	while (list != NULL) {
		interface = list->data;
		g_object_unref(interface);

		list = list->next;
	}
		
	g_list_free(list1);

	return(ret);
}

/**
 * lql_con_get_interfaces:
 * @self: The LQLCon object to use.
 *
 * This function returns a GList containing LQLInterface
 * objects representing each interface in the system.
 *
 * Returns: A GList containing LQLInterface objects.
 */
GList*
lql_con_get_interfaces(LQLCon *self)
{
	g_assert(self != NULL);

	return(lql_con_helper_get_interfaces(self));
}

/**
 * lql_con_get_qdiscs:
 * @self: The LQLCon object to use.
 * @interface: The interface to look at.
 *
 * This function returns a GList containing LQLQDisc
 * objects representing each queueing discipiline in the system.
 *
 * Returns: A GList containing LQLQDisc objects.
 */
GList*
lql_con_get_qdiscs(LQLCon *self,
		   LQLInterface *interface)
{
	g_assert(self != NULL);
	g_assert(interface != NULL);

	return(lql_con_helper_get_qdiscs(self, lql_interface_get_ifIndex(interface)));
}

/**
 * lql_con_get_classes:
 * @self: The LQLCon object to use.
 * @interface: The LQLInterface for which to get classes.
 *
 * This function returns a GList containing LQLClass
 * objects representing each class in the system.
 *
 * Returns: A GList containing LQLClass objects.
 */
GList*
lql_con_get_classes(LQLCon *self,
		    LQLInterface *interface)
{
	g_assert(self != NULL);
	g_assert(interface != NULL);

	return(lql_con_helper_get_classes(self, lql_interface_get_ifIndex(interface)));
}

/**
 * lql_con_get_classifiers:
 * @self: The LQLCon object to use.
 * @interface: The LQLInterface representing the interface to get classifiers for.
 *
 * This function returns a GList containing LQLClassifier
 * objects representing each classifier in the system.
 *
 * Returns: A GList containing LQLClassifier objects.
 */
GList*
lql_con_get_classifiers(LQLCon *self,
			LQLInterface *interface)
{
	g_assert(self != NULL);
	g_assert(interface != NULL);

	return(lql_con_helper_get_classifiers(self, lql_interface_get_ifIndex(interface)));
}

/**
 * lql_con_get_cell_log:
 * @self: The LQLCon object to use.
 * @size: The size of the packet.
 *
 * This function figures out how many right shifts are
 * needed to get the packet size below 256. Classes need this
 * value when setting up rate tables.
 *
 * Returns: The number of right shift to get the size below 256.
 **/
int
lql_con_get_cell_log(LQLCon *self,
		     int size)
{
	return(lql_con_helper_get_cell_log(size));
}

/**
 * lql_con_calc_xmit_time:
 * @self: The LQLCon object to use.
 * @bps: The speed to use in the calculation.
 * @size: The size of the packet to send.
 *
 * This function figures out how long it would take to send
 * size bytes at (bps) bytes per second.
 *
 * Returns: The the number of ticks it would take to send the data.
 **/
unsigned
lql_con_calc_xmit_time(LQLCon *self,
		       guint32 bps,
		       guint32 size)
{
	return(lql_con_helper_calc_xmit_time(self, bps, size));
}

/**
 * lql_con_calc_rate_table:
 * @self: The LQLCon object to use.
 * @bps: The speed in bytes per second.
 * @rtab: A pointer to the table to fill out.
 * @mtu: The MTU size of the interface.
 * @mpu: The MPU size of the interface.
 *
 * This function fills out the rate tables. This is needed by
 * some classes.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_con_calc_rate_table(LQLCon *self,
			guint32 bps,
			guint32 *rtab,
			guint32 mtu,
			guint32  mpu)
{
	return(lql_con_helper_calc_rate_table(self, bps,rtab, mtu, mpu));
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_CON_SEQUENCE = 1,
        LQL_CON_TICKINUSEC = 2,
        LQL_CON_HZ = 3,
};

static void
lql_con_instance_init(GTypeInstance *instance,
		      gpointer g_class)
{
        LQLCon *self = (LQLCon *)instance;

        self->private = g_new0 (LQLConPrivate, 1);
        self->private->dispose_has_run = FALSE;

	if (!lql_con_helper_init(self)) {
		g_printerr("LQLCon: Could not init.\n");
	}
}

static void
lql_con_set_property(GObject *object,
		     guint property_id,
		     const GValue *value,
		     GParamSpec *pspec)
{
        // LQLCon *self = (LQLCon *) object;

        switch (property_id) {
        default:
                g_assert (FALSE);
                break;
        }
}

static void
lql_con_get_property(GObject *object,
		     guint property_id,
		     GValue *value,
		     GParamSpec *pspec)
{
        LQLCon *self = (LQLCon *) object;

        switch (property_id) {
        case LQL_CON_SEQUENCE: {
                g_value_set_int(value, self->private->seq);
                break;
        }
        case LQL_CON_TICKINUSEC: {
                g_value_set_double(value, self->private->tickInUsec);
                break;
        }
        case LQL_CON_HZ: {
                g_value_set_double(value, self->private->hz);
                break;
        }
        default:
                g_assert (FALSE);
                break;
        }
}

static GObject *
lql_con_constructor(GType type,
		    guint n_construct_properties,
		    GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                LQLConClass *klass;
                GObjectClass *parent_class;  
                klass = LQL_CON_CLASS(g_type_class_peek(LQL_TYPE_CON));
                parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
                obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */
	

        return obj;
}

static void
lql_con_dispose (GObject *obj)
{
        LQLCon *self = (LQLCon *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_con_finalize (GObject *obj)
{
        LQLCon *self = (LQLCon *)obj;

	/* Free instance variables if necessary. */
	if (!nl_close(self->private->con)) {
		g_printerr("Could not close netlink connection.\n");
	}
        
	/* Free the private struct. */
        g_free(self->private);
}


static void
lql_con_class_init (gpointer g_class,
		    gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        //LQLConClass *klass = LQL_CON_CLASS (g_class);
        GParamSpec *pspec;

        gobject_class->set_property = lql_con_set_property;
        gobject_class->get_property = lql_con_get_property;
        gobject_class->dispose = lql_con_dispose;
        gobject_class->finalize = lql_con_finalize;
        gobject_class->constructor = lql_con_constructor;

	/* Set the virtual methods. */

	/* Setup the properties. */
        pspec = g_param_spec_int("sequence",
                                 "LQLCon sequence number",
                                 "Get the next sequence number for communicating with the"
				 " kernel",
                                 0 /* minimum value */,
                                 4000000 /* maximum value */,
                                 0 /* default value */,
                                 G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CON_SEQUENCE,
                                        pspec);

        pspec = g_param_spec_double("tick-in-usec",
				    "The number of ticks in a usec",
				    "Set/Get the number of ticks in a usec",
				    0 /* minimum value */,
				    4000000 /* maximum value */,
				    0 /* default value */,
				    G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CON_TICKINUSEC,
                                        pspec);

        pspec = g_param_spec_double("hz",
				    "The system HZ setting",
				    "Get the system HZ setting",
				    0 /* minimum value */,
				    4000000 /* maximum value */,
				    0 /* default value */,
				    G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_CON_HZ,
                                        pspec);
}

GType lql_con_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLConClass),
                        NULL, /* base_init */
                        NULL, /* base_finalize */
                        lql_con_class_init, /* class_init */
                        NULL, /* class_finalize */
                        NULL, /* class_data */
                        sizeof (LQLCon),
                        0, /* n_preallocs */
                        lql_con_instance_init /* instance_init */
                };
                type = g_type_register_static(G_TYPE_OBJECT,
					      "LQLCon",
					      &info, 0);
        }
        return type;
}
