/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-element.h"
#include "lql-element-private.h"

/********************************************************************
 * Non-Virtual API
 ********************************************************************/

/**
 * lql_element_new:
 *
 * Build a new instance of LQLElement. 
 * 
 * Return value: A new LQLElement instance.
 **/
LQLElement*
lql_element_new(void)
{
        return(LQL_ELEMENT(g_object_new(LQL_TYPE_ELEMENT, NULL)));
}

/**
 * lql_element_set_interface:
 * @self: The LQLElement instance.
 * @interface: An LQLInterface instance.
 * 
 * Set the interface that this LQLElement is attached to.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_element_set_interface(LQLElement *self,
			  LQLInterface *interface)
{
	int index = 0;

	g_assert(self != NULL);
	g_assert(interface != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	index = lql_interface_get_ifIndex(LQL_INTERFACE(interface));
	
	lql_element_set_ifIndex(LQL_ELEMENT(self), index);

	return(TRUE);
}

/**
 * lql_element_get_ifIndex:
 * @self: The LQLElement instance.
 * 
 * Get the interface index this LQLElement is attached to.
 * 
 * Return value: The element index, -1 on error.
 **/
int
lql_element_get_ifIndex(LQLElement *self)
{
        int ifIndex = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }

        g_object_get(LQL_ELEMENT(self), "interface-index", &ifIndex, NULL);

        return(ifIndex);
}

/**
 * lql_element_set_ifIndex:
 * @self: The LQLElement instance.
 * @ifIndex: The index of a network interface.
 * 
 * Set the network interface index that this Element is attached to. Most of the time
 * you will want to use lql_element_set_interface() instead of the _ifIndex() functions.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_element_set_ifIndex(LQLElement *self,
			int ifIndex)
{
        g_assert(self != NULL);
        g_assert(ifIndex > 0);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_ELEMENT(self), "interface-index", ifIndex, NULL);
        
        return(TRUE);
}

/**
 * lql_element_get_parentHandle:
 * @self: The LQLElement instance.
 * 
 * Get the parentHandle of this Element.
 * 
 * Return value: The element parentHandle, -1 on error.
 **/
guint32
lql_element_get_parentHandle(LQLElement *self)
{
        guint32 parentHandle = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_ELEMENT(self), "parent-handle", &parentHandle, NULL);

        return(parentHandle);
}

/**
 * lql_element_set_parentHandle:
 * @self: The LQLElement instance.
 * @parentHandle: The index to set the parentHandle instance variable to.
 * 
 * Set the parentHandle field of this object.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_element_set_parentHandle(LQLElement *self,
			     guint32 parentHandle)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_ELEMENT(self), "parent-handle", parentHandle, NULL);
        
        return(TRUE);
}

/**
 * lql_element_get_handle:
 * @self: The LQLElement instance.
 * 
 * Get the handle of this object.
 * 
 * Return value: The element HANDLE.
 **/
guint32
lql_element_get_handle(LQLElement *self)
{
        guint32 handle = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_ELEMENT(self), "handle", &handle, NULL);

        return(handle);
}

/**
 * lql_element_set_handle:
 * @self: The LQLElement instance.
 * @handle: The value to set the handle instance variable to.
 * 
 * Set the handle field of this object.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_element_set_handle(LQLElement *self,
		       guint32 handle)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_ELEMENT(self), "handle", handle, NULL);
        
        return(TRUE);
}

/**
 * lql_element_fill_from_nlhdr:
 * @self: The LQLElment instance.
 * @hdr: The nlmsghdr from the kernel.
 * 
 * This function is used by the implementations of sub-classes to fill out the LQLElement
 * specific fields from the data returned by the kernel. There is no reason for a normal
 * application to use this function.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_element_fill_from_nlhdr(LQLElement *self,
			    const struct nlmsghdr *hdr)
{
	g_assert(self != NULL);
	g_assert(hdr != NULL);

	/* If the parent comes back as 0 set it to TC_H_ROOT so future deletes and
	   modifications work properly. */
	if (nl_tcmsg_get_parent(hdr) == 0) {
		lql_element_set_parentHandle(LQL_ELEMENT(self), TC_H_ROOT);
	} else {
		lql_element_set_parentHandle(LQL_ELEMENT(self), nl_tcmsg_get_parent(hdr));
	}
	lql_element_set_handle(LQL_ELEMENT(self), nl_tcmsg_get_handle(hdr));
	lql_element_set_ifIndex(LQL_ELEMENT(self), nl_tcmsg_get_ifindex(hdr));
	
	return(TRUE);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_ELEMENT_IFINDEX = 1,
	LQL_ELEMENT_HANDLE = 2,
	LQL_ELEMENT_PARENTHANDLE = 3,
};

static void
lql_element_instance_init(GTypeInstance *instance,
			  gpointer g_class)
{
        LQLElement *self = (LQLElement *)instance;
        self->private = g_new0(LQLElementPrivate, 1);
        self->private->dispose_has_run = FALSE;
}

static void
lql_element_set_property(GObject *object,
			 guint property_id,
			 const GValue *value,
			 GParamSpec *pspec)
{
        LQLElement *self = (LQLElement *) object;

        switch (property_id) {
        case LQL_ELEMENT_IFINDEX: 
        {
                self->private->ifIndex = g_value_get_int(value);
                break;
        }
	case LQL_ELEMENT_HANDLE:
	{
		self->private->handle = g_value_get_uint(value);
		break;
	}
	case LQL_ELEMENT_PARENTHANDLE:
	{
		self->private->parentHandle = g_value_get_uint(value);
		break;
	}
        default:
                g_assert (FALSE);
                break;
        }
}

static void
lql_element_get_property(GObject *object,
			 guint property_id,
			 GValue *value,
			 GParamSpec *pspec)
{
        LQLElement *self = (LQLElement *) object;

        switch (property_id) {
        case LQL_ELEMENT_IFINDEX: 
        {
                g_value_set_int(value, self->private->ifIndex);
                break;
        }
	case LQL_ELEMENT_HANDLE:
	{
		g_value_set_uint(value, self->private->handle);
		break;
	}
	case LQL_ELEMENT_PARENTHANDLE:
	{
		g_value_set_uint(value, self->private->parentHandle);
		break;
	}
        default:
                g_assert (FALSE);
                break;
        }
}

static GObject *
lql_element_constructor(GType type,
			guint n_construct_properties,
			GObjectConstructParam *construct_properties)
{
        GObject *obj = NULL;

        {
                /* Invoke parent constructor. */
                LQLElementClass *klass;
                GObjectClass *parent_class;  
                klass = LQL_ELEMENT_CLASS(g_type_class_peek(LQL_TYPE_ELEMENT));
                parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
                obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_element_dispose (GObject *obj)
{
        LQLElement *self = (LQLElement *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_element_finalize (GObject *obj)
{
        LQLElement *self = (LQLElement *)obj;

        /*
         * Here, complete object destruction.
         * You might not need to do much...
         */

        
        g_free (self->private);
}

static void
lql_element_class_init (gpointer g_class,
			gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        //LQLElementClass *klass = LQL_ELEMENT_CLASS(g_class);
        GParamSpec *pspec = NULL;

        gobject_class->set_property = lql_element_set_property;
        gobject_class->get_property = lql_element_get_property;
        gobject_class->dispose = lql_element_dispose;
        gobject_class->finalize = lql_element_finalize;
        gobject_class->constructor = lql_element_constructor;

	/* Set the virtual methods. */

	/* Setup the properties. */
        pspec = g_param_spec_int("interface-index",
                                 "Network interface index",
                                 "Set/Get the network interface index that this element is"
				 " attached to",
                                 G_MININT /* minimum value */,
                                 G_MAXINT /* maximum value */,
                                 0 /* default value */,
                                 G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_ELEMENT_IFINDEX,
                                        pspec);

        pspec = g_param_spec_uint("handle",
				  "The 32bit unsigned integer that "
				  "identifies this element to the kernel",
				  "Set/Get the 32bit unsigned integer that "
				  "identifies this element to the kernel",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_ELEMENT_HANDLE,
                                        pspec);

        pspec = g_param_spec_uint("parent-handle",
				  "The 32bit unsigned integer that "
				  "identifies this elements parent element",
				  "Set/Get the 32bit unsigned integer that "
				  "identifies this elements parent element",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_ELEMENT_PARENTHANDLE,
                                        pspec);
}

GType lql_element_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLElementClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_element_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLElement),
                        0,      /* n_preallocs */
                        lql_element_instance_init    /* instance_init */
                };
                type = g_type_register_static(G_TYPE_OBJECT,
					      "LQLElement",
					      &info, 0);
        }

        return type;
}
