/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-interface.h"
#include "lql_interface_helper.h"

struct _LQLInterfacePrivate
{
        gboolean dispose_has_run;

        gchar *name;
        int ifIndex;
        int mtu;
};

/********************************************************************
 * Non-Virutal API
 ********************************************************************/

/**
 * lql_interface_new:
 *
 * Build a new instance of LQLInterface. 
 * 
 * Return value: A new LQLInterface instance.
 **/
LQLInterface*
lql_interface_new(void)
{
        return(LQL_INTERFACE(g_object_new(LQL_TYPE_INTERFACE, NULL)));
}

/**
 * lql_interface_fill_from_nlhdr:
 * @self: The LQLInterface object to work with.
 * @hdr: The interface information from the kernel.
 * 
 * Fill in the instance variables for this instance of LQLInterface
 * to match the network interface information in hdr.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_interface_fill_from_nlhdr(LQLInterface *self,
                              const struct nlmsghdr *hdr)
{
        g_assert(self != NULL);
        g_assert(hdr != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }

        return(lql_interface_helper_fill_from_nlhdr(LQL_INTERFACE(self), hdr));
}

/**
 * lql_interface_get_ifIndex:
 * @self: The LQLInterface object to work with.
 * 
 * Get the index of the interface that this LQLInterface represents.
 * 
 * Return value: The interface index, -1 on error.
 **/
int
lql_interface_get_ifIndex(LQLInterface *self)
{
        int ifIndex = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_INTERFACE(self), "ifIndex", &ifIndex, NULL);

        return(ifIndex);
}

/**
 * lql_interface_set_ifIndex:
 * @self: The LQLInterface object to work with.
 * @ifIndex: The index to set the ifIndex instance variable to.
 * 
 * Set the ifIndex field of this object.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_interface_set_ifIndex(LQLInterface *self,
                          int ifIndex)
{
        g_assert(self != NULL);
        g_assert(ifIndex > 0);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_INTERFACE(self), "ifIndex", ifIndex, NULL);
        
        return(TRUE);
}

/**
 * lql_interface_get_mtu:
 * @self: The LQLInterface object to work with.
 * 
 * Get the MTU.
 * 
 * Return value: The value of the MTU instance variable, -1 on error.
 **/
int
lql_interface_get_mtu(LQLInterface *self)
{
        int mtu = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(-1);
        }
        
        g_object_get(LQL_INTERFACE(self), "mtu", &mtu, NULL);

        return(mtu);
}

/**
 * lql_interface_set_mtu:
 * @self: The LQLInterface object to work with.
 * @mtu: The value to set the MTU field to.
 * 
 * Set the MTU instance variable.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_interface_set_mtu(LQLInterface *self,
                      int mtu)
{
        g_assert(self != NULL);
        g_assert(mtu > 0);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_INTERFACE(self), "mtu", mtu, NULL);
        
        return(TRUE);
}

/**
 * lql_interface_get_name:
 * @self: The LQLInterface object to work with.
 * 
 * Get the name of the interface.
 * 
 * Return value: The name of this interface, NULL on error.
 **/
gchar *
lql_interface_get_name(LQLInterface *self)
{
        gchar *name = NULL;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(NULL);
        }
        
        g_object_get(LQL_INTERFACE(self), "name", &name, NULL);

        return(name);
}

/**
 * lql_interface_set_name:
 * @self: The LQLInterface object to work with.
 * @name: A pointer to the string to set the name to.
 * 
 * Set the name instance variable.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_interface_set_name(LQLInterface *self,
                       const gchar *name)
{
        g_assert(self != NULL);
        g_assert(name != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_INTERFACE(self), "name", name, NULL);
        
        return(TRUE);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_INTERFACE_NAME = 1,
        LQL_INTERFACE_MTU = 2,
        LQL_INTERFACE_IFINDEX = 3,
};

static void
lql_interface_instance_init(GTypeInstance *instance,
			    gpointer g_class)
{
        LQLInterface *self = (LQLInterface *)instance;
        self->private = g_new0 (LQLInterfacePrivate, 1);
        self->private->dispose_has_run = FALSE;
}


static void
lql_interface_set_property(GObject *object,
			   guint property_id,
			   const GValue *value,
			   GParamSpec *pspec)
{
        LQLInterface *self = (LQLInterface *) object;

        switch (property_id) {
        case LQL_INTERFACE_NAME: {
                g_free(self->private->name);
                self->private->name = g_value_dup_string(value);
                break;
        }
        case LQL_INTERFACE_MTU: {
                self->private->mtu = g_value_get_int(value);
                break;
        }
        case LQL_INTERFACE_IFINDEX: 
        {
                self->private->ifIndex = g_value_get_int(value);
                break;
        }
        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static void
lql_interface_get_property(GObject *object,
                           guint property_id,
                           GValue *value,
                           GParamSpec *pspec)
{
        LQLInterface *self = (LQLInterface *) object;

        switch (property_id) {
        case LQL_INTERFACE_NAME: {
                g_value_set_string(value, self->private->name);
                break;
        }
        case LQL_INTERFACE_MTU: {
                g_value_set_int(value, self->private->mtu);
                break;
        }
        case LQL_INTERFACE_IFINDEX: 
        {
                g_value_set_int(value, self->private->ifIndex);
                break;
        }
        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static GObject *
lql_interface_constructor(GType type,
                          guint n_construct_properties,
                          GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                LQLInterfaceClass *klass;
                GObjectClass *parent_class;  
                klass = LQL_INTERFACE_CLASS (g_type_class_peek (LQL_TYPE_INTERFACE));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_interface_dispose (GObject *obj)
{
        LQLInterface *self = (LQLInterface *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_interface_finalize (GObject *obj)
{
        LQLInterface *self = (LQLInterface *)obj;

        /*
         * Here, complete object destruction.
         * You might not need to do much...
         */
        g_free (self->private->name);
        
        g_free (self->private);
}


static void
lql_interface_class_init (gpointer g_class,
			  gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
        //LQLInterfaceClass *klass = LQL_INTERFACE_CLASS (g_class);
        GParamSpec *pspec;

        gobject_class->set_property = lql_interface_set_property;
        gobject_class->get_property = lql_interface_get_property;
        gobject_class->dispose = lql_interface_dispose;
        gobject_class->finalize = lql_interface_finalize;
        gobject_class->constructor = lql_interface_constructor;

	/* Set the virtual methods. */

	/* Setup the properties. */
        pspec = g_param_spec_string("name",
                                    "Network interface name",
                                    "Set/Get LQLInterface name",
                                    "no-name-set" /* default value */,
                                    G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_INTERFACE_NAME,
                                        pspec);

        pspec = g_param_spec_int("mtu",
                                 "Network Interface MTU",
                                 "Set/Get LQLInterface MTU",
                                 0 /* minimum value */,
                                 G_MAXINT /* maximum value */,
                                 0 /* default value */,
                                 G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_INTERFACE_MTU,
                                        pspec);

        pspec = g_param_spec_int("ifIndex",
                                 "Network interface index",
                                 "Set/Get LQLInterface index",
                                 0 /* minimum value */,
                                 G_MAXINT /* maximum value */,
                                 0 /* default value */,
                                 G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_INTERFACE_IFINDEX,
                                        pspec);
}

GType lql_interface_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLInterfaceClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_interface_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLInterface),
                        0,      /* n_preallocs */
                        lql_interface_instance_init    /* instance_init */
                };
                type = g_type_register_static (G_TYPE_OBJECT,
                                               "LQLInterface",
                                               &info, 0);
        }
        return type;
}
