/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-qdisc-dsmark.h"
#include "lql-qdisc-dsmark-private.h"
#include "lql_qdisc_dsmark_helper.h"

/********************************************************************
 * Non-Virutal API
 ********************************************************************/

/**
 * lql_qdisc_dsmark_new:
 *
 * Build a new instance of LQLQDiscDSMark. 
 * 
 * Return value: A new LQLQDiscDSMark instance.
 **/
LQLQDiscDSMark*
lql_qdisc_dsmark_new(void)
{
        return(LQL_QDISC_DSMARK(g_object_new(LQL_TYPE_QDISC_DSMARK, NULL)));
}

/**
 * lql_qdisc_dsmark_get_indices:
 * @self: The LQLQDiscDSMark instance.
 * 
 * Get the size of the mask/value table.
 * 
 * Return value: The size of the mask/value table.
 **/
guint16
lql_qdisc_dsmark_get_indices(LQLQDiscDSMark *self)
{
	guint32 value;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
	
	g_object_get(LQL_QDISC_DSMARK(self), "indices", &value, NULL);
	
	return(value);
}

/**
 * lql_qdisc_dsmark_set_indices:
 * @self: The LQLQDiscDSMark instance.
 * @indices: 
 * 
 * Set the size of the mask/value table.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_dsmark_set_indices(LQLQDiscDSMark *self,
			     guint16 indices)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_DSMARK(self), "indices", indices, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_dsmark_get_setTCIndex:
 * @self: The LQLQDiscDSMark instance.
 * 
 * Get whether the QDisc is currently setting the DS value into skb->tc_index.
 * 
 * Return value: The current value of setTCIndex.
 **/
gboolean
lql_qdisc_dsmark_get_setTCIndex(LQLQDiscDSMark *self)
{
	gboolean value;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
	
	g_object_get(LQL_QDISC_DSMARK(self), "set-tc-index", &value, NULL);
	
	return(value);
}

/**
 * lql_qdisc_dsmark_set_setTCIndex:
 * @self: The LQLQDiscDSMark instance.
 * @setTCIndex: TRUE to set skb->tc_index, FALSE to not.
 * 
 * Set whether or not the QDisc is currently setting the DS value into skb->tc_index.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_dsmark_set_setTCIndex(LQLQDiscDSMark *self,
				gboolean setTCIndex)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_DSMARK(self), "set-tc-index", setTCIndex, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_dsmark_get_defaultIndex:
 * @self: The LQLQDiscDSMark instance.
 * 
 * Get the current default index setting.
 * 
 * Return value: The default index.
 **/
guint16
lql_qdisc_dsmark_get_defaultIndex(LQLQDiscDSMark *self)
{
	guint32 value;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
	
	g_object_get(LQL_QDISC_DSMARK(self), "default-index", &value, NULL);
	
	return(value);
}

/**
 * lql_qdisc_dsmark_set_defaultIndex:
 * @self: The LQLQDiscDSMark instance.
 * @defaultIndex: 
 * 
 * Set the current default index setting.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_dsmark_set_defaultIndex(LQLQDiscDSMark *self,
				  guint16 defaultIndex)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_DSMARK(self), "default-index", defaultIndex, NULL);
        
        return(TRUE);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/**
 * lql_qdisc_dsmark_fill_from_nlhdr:
 * @self: The LQLQDisc object to work with.
 * @hdr: The nl packet to get the queue information from.
 *
 * This method takes a nlmsg from the kernel and set this
 * objects instance variables to match it. This is used
 * to create objects that match the ones currently active
 * in the system. This overrides the default LQLQDisc
 * implementation of this method.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_dsmark_fill_from_nlhdr(LQLQDisc *self,
				 const struct nlmsghdr *hdr)
{
	g_assert(self != NULL);
	g_assert(hdr != NULL);

	return(lql_qdisc_dsmark_helper_fill_from_nlhdr(LQL_QDISC_DSMARK(self), hdr));
}

/**
 * lql_qdisc_dsmark_add:
 * @self: The LQLQDisc object to add to the system.
 * @con: The LQLCon connection to use when adding this queueing discipline.
 *
 * Add this DSMark queueing discipline to the system. This overrides
 * the default implementation of this function in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_dsmark_add(LQLQDisc *self,
		     LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_dsmark_helper_add(LQL_QDISC_DSMARK(self), con));
}

/**
 * lql_qdisc_dsmark_modify:
 * @self: The LQLQDisc object to modify.
 * @con: The LQLCon connection to use when adding this qdisc.
 *
 * Modify the DSMark queueing discipline in the system to match the current
 * object state. This method overrides the default
 * implementation in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_dsmark_modify(LQLQDisc *self,
		     LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_dsmark_helper_modify(LQL_QDISC_DSMARK(self), con));
}

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_QDISC_DSMARK_INDICES = 1,
        LQL_QDISC_DSMARK_DEFAULT_INDEX = 2,
	LQL_QDISC_DSMARK_SET_TC_INDEX = 3,
};

static void
lql_qdisc_dsmark_instance_init(GTypeInstance *instance,
			    gpointer g_class)
{
	LQLQDiscDSMark *self = (LQLQDiscDSMark *)instance;
	self->private = g_new0(LQLQDiscDSMarkPrivate, 1);
	self->private->dispose_has_run = FALSE;

	self->private->indices = 64;
	self->private->setTCIndex = TRUE;
	self->private->defaultIndex = 1;
}

static void
lql_qdisc_dsmark_set_property(GObject *object,
			   guint property_id,
			   const GValue *value,
			   GParamSpec *pspec)
{
        LQLQDiscDSMark *self = (LQLQDiscDSMark *) object;

        switch (property_id) {
        case LQL_QDISC_DSMARK_INDICES:
        {
                self->private->indices = g_value_get_uint(value);
                break;
        }
        case LQL_QDISC_DSMARK_SET_TC_INDEX:
        {
                self->private->setTCIndex = g_value_get_boolean(value);
                break;
        }
        case LQL_QDISC_DSMARK_DEFAULT_INDEX:
        {
                self->private->defaultIndex = g_value_get_uint(value);
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static void
lql_qdisc_dsmark_get_property(GObject *object,
                           guint property_id,
                           GValue *value,
                           GParamSpec *pspec)
{
        LQLQDiscDSMark *self = (LQLQDiscDSMark *) object;

        switch (property_id) {
        case LQL_QDISC_DSMARK_INDICES:
        {
                g_value_set_uint(value, self->private->indices);
                break;
        }
        case LQL_QDISC_DSMARK_SET_TC_INDEX: 
        {
                g_value_set_boolean(value, self->private->setTCIndex);
                break;
        }
        case LQL_QDISC_DSMARK_DEFAULT_INDEX:
        {
                g_value_set_uint(value, self->private->defaultIndex);
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static GObject *
lql_qdisc_dsmark_constructor(GType type,
                          guint n_construct_properties,
                          GObjectConstructParam *construct_properties)
{
        GObject *obj = NULL;

        {
                /* Invoke parent constructor. */
		LQLQDiscDSMarkClass *klass;
		GObjectClass *parent_class;  
		klass = LQL_QDISC_DSMARK_CLASS(g_type_class_peek(LQL_TYPE_QDISC_DSMARK));
		parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
		obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_qdisc_dsmark_dispose(GObject *obj)
{
        LQLQDiscDSMark *self = (LQLQDiscDSMark *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_qdisc_dsmark_finalize(GObject *obj)
{
        LQLQDiscDSMark *self = (LQLQDiscDSMark *)obj;

        /*
         * Here, complete object destruction.
         */
        
        g_free (self->private);
}


static void
lql_qdisc_dsmark_class_init(gpointer g_class,
			 gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        LQLQDiscClass *klass = LQL_QDISC_CLASS(g_class);
        GParamSpec *pspec = NULL;

        gobject_class->set_property = lql_qdisc_dsmark_set_property;
        gobject_class->get_property = lql_qdisc_dsmark_get_property;
        gobject_class->dispose = lql_qdisc_dsmark_dispose;
        gobject_class->finalize = lql_qdisc_dsmark_finalize;
        gobject_class->constructor = lql_qdisc_dsmark_constructor;

	/* Set the virtual methods. */
	klass->fill_from_nlhdr = lql_qdisc_dsmark_fill_from_nlhdr;
	klass->add = lql_qdisc_dsmark_add;
	klass->modify = lql_qdisc_dsmark_modify;
	klass->print = lql_qdisc_dsmark_helper_print;

	/* Setup the properties. */
        pspec = g_param_spec_uint("indices",
				  "The size of the mask/value table.",
				  "Set/Get the size of the mask/value table.",
				  0 /* minimum value */,
				  65535 /* maximum value */,
				  64 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_DSMARK_INDICES,
                                        pspec);

        pspec = g_param_spec_boolean("set-tc-index",
				     "Whether or not the kernel places the DS value into"
				     " skb->tc_index",
				     "Set/Get whether or not the kernel places the DS value"
				     " into skb->tc_index",
				     TRUE /* default value */,
				     G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_DSMARK_SET_TC_INDEX,
                                        pspec);

        pspec = g_param_spec_uint("default-index",
				  "The default index into the mask/value table.",
				  "Set/Get the default index into the mask/value table.",
				  0 /* minimum value */,
				  G_MAXINT /* maximum value */,
				  1 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_DSMARK_DEFAULT_INDEX,
                                        pspec);
}

GType lql_qdisc_dsmark_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLQDiscDSMarkClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_qdisc_dsmark_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLQDiscDSMark),
                        0,      /* n_preallocs */
                        lql_qdisc_dsmark_instance_init    /* instance_init */
                };
                type = g_type_register_static (LQL_TYPE_QDISC,
                                               "LQLQDiscDSMark",
                                               &info, 0);
        }
        return type;
}
