/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-qdisc-netem.h"
#include "lql-qdisc-netem-private.h"
#include "lql_qdisc_netem_helper.h"

/********************************************************************
 * Non-Virutal API
 ********************************************************************/

/**
 * lql_qdisc_netem_new:
 *
 * Build a new instance of LQLQDiscNetem. 
 * 
 * Return value: A new LQLQDiscNetem instance.
 **/
LQLQDiscNetem*
lql_qdisc_netem_new(void)
{
        return(LQL_QDISC_NETEM(g_object_new(LQL_TYPE_QDISC_NETEM, NULL)));
}

/**
 * lql_qdisc_netem_get_latency:
 * @self: The LQLQDiscNetem instance.
 * 
 * Return the amount of latency in ms that the netem qdisc is adding to packets.
 * 
 * Return value: The amount of latency being added to packets in ms
 **/
guint32
lql_qdisc_netem_get_latency(LQLQDiscNetem *self)
{
        guint32 latency = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_NETEM(self), "latency", &latency, NULL);

        return(latency);
}

/**
 * lql_qdisc_netem_set_latency:
 * @self: The LQLQDiscNetem instance.
 * @latency: The amount of latency in ms to add to packets.
 * 
 * Set the amount of latency in ms that the netem qdisc adds to packets.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_netem_set_latency(LQLQDiscNetem *self,
			    guint32 latency)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_NETEM(self), "latency", latency, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_netem_get_loss:
 * @self: The LQLQDiscNetem instance.
 * 
 * Get the current percentage of packets being dropped by this qdisc. This
 * is expressed as a float value between 0 and 100.
 * 
 * Return value: The percentage of packets being dropped.
 **/
float
lql_qdisc_netem_get_loss(LQLQDiscNetem *self)
{
        float loss = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_NETEM(self), "loss", &loss, NULL);

        return(loss);
}

/**
 * lql_qdisc_netem_set_loss:
 * @self: The LQLQDiscNetem instance.
 * @loss: The percentage of packets to drop.
 * 
 * Set the current percentage of packets being dropped by this qdisc. This
 * is expressed as a float value between 0 and 100.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_netem_set_loss(LQLQDiscNetem *self,
			 float loss)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_NETEM(self), "loss", loss, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_netem_get_limit:
 * @self: The LQLQDiscNetem instance.
 * 
 * Get the maximum number of packets this Netem instance will hold. The default
 * value is 1000. In order to change latency Netem must hold onto packets. This value
 * puts a upper limit on the number of packets held.
 * 
 * Return value: The maximum number of packets.
 **/
guint32
lql_qdisc_netem_get_limit(LQLQDiscNetem *self)
{
        guint32 limit = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_NETEM(self), "limit", &limit, NULL);

        return(limit);
}

/**
 * lql_qdisc_netem_set_limit:
 * @self: The LQLQDiscNetem instance.
 * @limit: The new packet limit.
 * 
 * Set the maximum number of packets that the Netem instance will hold.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_netem_set_limit(LQLQDiscNetem *self,
			  guint32 limit)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_NETEM(self), "limit", limit, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_netem_get_gap:
 * @self: The LQLQDiscNetem instance.
 * 
 * The gap value tells the Netem instance which packets are effected by the Netem
 * parameters. For example, gap=5 means the 5th, 10th, 15th ... packets are effected
 * while the rest are not.
 * 
 * Return value: The current gap value.
 **/
guint32
lql_qdisc_netem_get_gap(LQLQDiscNetem *self)
{
        guint32 gap = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_NETEM(self), "gap", &gap, NULL);

        return(gap);
}

/**
 * lql_qdisc_netem_set_gap:
 * @self: The LQLQDiscNetem instance.
 * @gap: The gap value.
 * 
 * Set the current gap value.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_netem_set_gap(LQLQDiscNetem *self,
			guint32 gap)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_NETEM(self), "gap", gap, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_netem_get_duplicate:
 * @self: The LQLQDiscNetem instance.
 * 
 * The duplicate value is the percentage of packets that are duplicated. This is expressed as
 * a float number between 0 and 100.
 * 
 * Return value: The current duplication percentage.
 **/
float
lql_qdisc_netem_get_duplicate(LQLQDiscNetem *self)
{
        float duplicate = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_NETEM(self), "duplicate", &duplicate, NULL);

        return(duplicate);
}

/**
 * lql_qdisc_netem_set_duplicate:
 * @self: The LQLQDiscNetem instance.
 * @duplicate: The percentage of packets to duplicate (0.0 to 100.0).
 * 
 * Set the percentage of packets that are being duplicated.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_netem_set_duplicate(LQLQDiscNetem *self,
			      float duplicate)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_NETEM(self), "duplicate", duplicate, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_netem_get_jitter:
 * @self: The LQLQDiscNetem instance.
 * 
 * Get the amount of jitter (in ms) applied to packets.
 * 
 * Return value: The amount of packet jitter in ms.
 **/
guint32
lql_qdisc_netem_get_jitter(LQLQDiscNetem *self)
{
        guint32 jitter = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_NETEM(self), "jitter", &jitter, NULL);

        return(jitter);
}

/**
 * lql_qdisc_netem_set_jitter:
 * @self: The LQLQDiscNetem instance.
 * @jitter: The amount of jitter in ms.
 * 
 * Set the amount of jitter (in ms) that is applied to packets.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_netem_set_jitter(LQLQDiscNetem *self,
			   guint32 jitter)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_NETEM(self), "jitter", jitter, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_netem_get_latencyCorrelation:
 * @self: The LQLQDiscNetem instance.
 * 
 * Network delay variation isn't purely random, so to emulate this there is a
 * latency correlation value. This correlation is expressed as a precentage.
 * For example, a value of 25% means the next packets latency is 25% dependant on 
 * the previous packet.
 * 
 * Return value: The current latency correlation value expressed as a precentage.
 **/
float
lql_qdisc_netem_get_latencyCorrelation(LQLQDiscNetem *self)
{
        float latency = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_NETEM(self), "latency-correlation", &latency, NULL);

        return(latency);
}

/**
 * lql_qdisc_netem_set_latencyCorrelation:
 * @self: The LQLQDiscNetem instance.
 * @latency: The latency correlation value expressed as a float between 0.0 and 100.0.
 * 
 * Set the current latency correlation value.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_netem_set_latencyCorrelation(LQLQDiscNetem *self,
				       float latency)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_NETEM(self), "latency-correlation", latency, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_netem_get_lossCorrelation:
 * @self: The LQLQDiscNetem instance.
 * 
 * Network loss variation isn't purely random, so to emulate this there is a
 * loss correlation value. This correlation is expressed as a precentage.
 * For example, a value of 25% means the next packets chance of being dropped is 25%
 * dependant on the previous packet.
 * 
 * Return value: 
 **/
float
lql_qdisc_netem_get_lossCorrelation(LQLQDiscNetem *self)
{
        float loss = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_NETEM(self), "loss-correlation", &loss, NULL);

        return(loss);
}

/**
 * lql_qdisc_netem_set_lossCorrelation:
 * @self: The LQLQDiscNetem instance.
 * @loss: The loss correlation value expressed as a float between 0.0 and 100.0.
 * 
 * Set the current loss correlation value.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_netem_set_lossCorrelation(LQLQDiscNetem *self,
				    float loss)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_NETEM(self), "loss-correlation", loss, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_netem_get_duplicateCorrelation:
 * @self: The LQLQDiscNetem instance.
 * 
 * Network duplication variation isn't purely random, so to emulate this there is a
 * duplication correlation value. This correlation is expressed as a precentage.
 * For example, a value of 25% means the next packets chance of being duplicated is 25%
 * dependant on the previous packet.
 * 
 * Return value: The current duplication correlation value expressed as a float (0.0 to 100.0).
 **/
float
lql_qdisc_netem_get_duplicateCorrelation(LQLQDiscNetem *self)
{
        float duplicate = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_NETEM(self), "duplicate-correlation", &duplicate, NULL);

        return(duplicate);
}

/**
 * lql_qdisc_netem_set_duplicateCorrelation:
 * @self: The LQLQDiscNetem instance.
 * @duplicate: The duplication correlation value expressed as a percentage (0.0% to 100.0%).
 * 
 * Set the duplication correlation value.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_netem_set_duplicateCorrelation(LQLQDiscNetem *self,
					 float duplicate)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_NETEM(self), "duplicate-correlation", duplicate, NULL);
        
        return(TRUE);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/**
 * lql_qdisc_netem_fill_from_nlhdr:
 * @self: The LQLQDisc object to work with.
 * @hdr: The nl packet to get the queue information from.
 *
 * This method takes a nlmsg from the kernel and set this
 * objects instance variables to match it. This is used
 * to create objects that match the ones currently active
 * in the system. This overrides the default LQLQDisc
 * implementation of this method.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_netem_fill_from_nlhdr(LQLQDisc *self,
				const struct nlmsghdr *hdr)
{
	g_assert(self != NULL);
	g_assert(hdr != NULL);

	return(lql_qdisc_netem_helper_fill_from_nlhdr(LQL_QDISC_NETEM(self), hdr));
}

/**
 * lql_qdisc_netem_add:
 * @self: The LQLQDisc object to add to the system.
 * @con: The LQLCon connection to use when adding this queueing discipline.
 *
 * Add this NETEM queueing discipline to the system. This overrides
 * the default implementation of this function in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_netem_add(LQLQDisc *self,
		    LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_netem_helper_add(LQL_QDISC_NETEM(self), con));
}

/**
 * lql_qdisc_netem_modify:
 * @self: The LQLQDisc object to modify.
 * @con: The LQLCon connection to use when adding this qdisc.
 *
 * Modify the Netem queueing discipline in the system to match the current
 * object state. This method overrides the default
 * implementation in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_netem_modify(LQLQDisc *self,
		       LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_netem_helper_modify(LQL_QDISC_NETEM(self), con));
}

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_QDISC_NETEM_LATENCY = 1,
        LQL_QDISC_NETEM_LIMIT = 2,
	LQL_QDISC_NETEM_LOSS = 3,
	LQL_QDISC_NETEM_GAP = 4,
	LQL_QDISC_NETEM_DUPLICATE = 5,
	LQL_QDISC_NETEM_JITTER = 6,
	LQL_QDISC_NETEM_LATENCY_CORR = 7,
	LQL_QDISC_NETEM_LOSS_CORR = 8,
	LQL_QDISC_NETEM_DUP_CORR = 9,
};

static void
lql_qdisc_netem_instance_init(GTypeInstance *instance,
			      gpointer g_class)
{
	LQLQDiscNetem *self = (LQLQDiscNetem *)instance;
	self->private = g_new0(LQLQDiscNetemPrivate, 1);
	self->private->dispose_has_run = FALSE;

	self->private->limit = 1000;
}

static void
lql_qdisc_netem_set_property(GObject *object,
			     guint property_id,
			     const GValue *value,
			     GParamSpec *pspec)
{
        LQLQDiscNetem *self = (LQLQDiscNetem *) object;

        switch (property_id) {
        case LQL_QDISC_NETEM_LATENCY:
        {
                self->private->latency = g_value_get_uint(value);
                break;
        }
        case LQL_QDISC_NETEM_LIMIT:
        {
                self->private->limit = g_value_get_uint(value);
                break;
        }
        case LQL_QDISC_NETEM_LOSS:
        {
                self->private->loss = g_value_get_float(value);
		if (self->private->loss > 100) {
			self->private->loss = 100;
		}
                break;
        }
        case LQL_QDISC_NETEM_GAP:
        {
                self->private->gap = g_value_get_uint(value);
                break;
        }
        case LQL_QDISC_NETEM_DUPLICATE:
        {
                self->private->duplicate = g_value_get_float(value);
		if (self->private->duplicate > 100) {
			self->private->duplicate = 100;
		}
                break;
        }
        case LQL_QDISC_NETEM_JITTER:
        {
                self->private->jitter = g_value_get_uint(value);
                break;
        }
        case LQL_QDISC_NETEM_LATENCY_CORR:
        {
                self->private->delayCorr = g_value_get_float(value);
		if (self->private->delayCorr > 100) {
			self->private->delayCorr = 100;
		}
                break;
        }
        case LQL_QDISC_NETEM_LOSS_CORR:
        {
                self->private->lossCorr = g_value_get_float(value);
		if (self->private->lossCorr > 100) {
			self->private->lossCorr = 100;
		}
                break;
        }
        case LQL_QDISC_NETEM_DUP_CORR:
        {
                self->private->dupCorr = g_value_get_float(value);
		if (self->private->dupCorr > 100) {
			self->private->dupCorr = 100;
		}
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static void
lql_qdisc_netem_get_property(GObject *object,
			     guint property_id,
			     GValue *value,
			     GParamSpec *pspec)
{
        LQLQDiscNetem *self = (LQLQDiscNetem *) object;

        switch (property_id) {
        case LQL_QDISC_NETEM_LATENCY: 
        {
                g_value_set_uint(value, self->private->latency);
                break;
        }
        case LQL_QDISC_NETEM_LIMIT:
        {
                g_value_set_uint(value, self->private->limit);
                break;
        }
        case LQL_QDISC_NETEM_LOSS:
        {
                g_value_set_float(value, self->private->loss);
                break;
        }
        case LQL_QDISC_NETEM_GAP:
        {
                g_value_set_uint(value, self->private->gap);
                break;
        }
        case LQL_QDISC_NETEM_DUPLICATE:
        {
                g_value_set_float(value, self->private->duplicate);
                break;
        }
        case LQL_QDISC_NETEM_JITTER:
        {
                g_value_set_uint(value, self->private->jitter);
                break;
        }
        case LQL_QDISC_NETEM_LATENCY_CORR:
        {
                g_value_set_float(value, self->private->delayCorr);
                break;
        }
        case LQL_QDISC_NETEM_LOSS_CORR:
        {
                g_value_set_float(value, self->private->lossCorr);
                break;
        }
        case LQL_QDISC_NETEM_DUP_CORR:
        {
                g_value_set_float(value, self->private->dupCorr);
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static GObject *
lql_qdisc_netem_constructor(GType type,
			    guint n_construct_properties,
			    GObjectConstructParam *construct_properties)
{
        GObject *obj = NULL;

        {
                /* Invoke parent constructor. */
		LQLQDiscNetemClass *klass;
		GObjectClass *parent_class;  
		klass = LQL_QDISC_NETEM_CLASS(g_type_class_peek(LQL_TYPE_QDISC_NETEM));
		parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
		obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_qdisc_netem_dispose(GObject *obj)
{
        LQLQDiscNetem *self = (LQLQDiscNetem *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_qdisc_netem_finalize(GObject *obj)
{
        LQLQDiscNetem *self = (LQLQDiscNetem *)obj;

        /*
         * Here, complete object destruction.
         */

        g_free (self->private);
}


static void
lql_qdisc_netem_class_init(gpointer g_class,
			   gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        LQLQDiscClass *klass = LQL_QDISC_CLASS(g_class);
        GParamSpec *pspec = NULL;

        gobject_class->set_property = lql_qdisc_netem_set_property;
        gobject_class->get_property = lql_qdisc_netem_get_property;
        gobject_class->dispose = lql_qdisc_netem_dispose;
        gobject_class->finalize = lql_qdisc_netem_finalize;
        gobject_class->constructor = lql_qdisc_netem_constructor;

	/* Set the virtual methods. */
	klass->fill_from_nlhdr = lql_qdisc_netem_fill_from_nlhdr;
	klass->add = lql_qdisc_netem_add;
	klass->modify = lql_qdisc_netem_modify;
	klass->print = lql_qdisc_netem_helper_print;
	klass->printStats = lql_qdisc_netem_helper_printStats;
	klass->updateStats = lql_qdisc_netem_helper_updateStats;

	/* Setup the properties. */
        pspec = g_param_spec_uint("latency",
				  "The amount of delay added to each packet in ms.",
				  "Set/Get the amount of delay added to each packet ms.",
				  0 /* minimum value */,
				  G_MAXUINT /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_NETEM_LATENCY,
                                        pspec);

        pspec = g_param_spec_uint("limit",
				  "The limit on the number of packets the QDisc will hold.",
				  "Set/Get the limit on the number of packets the QDisc will"
				  " hold",
				  1 /* minimum value */,
				  G_MAXUINT /* maximum value */,
				  1000 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_NETEM_LIMIT,
                                        pspec);

        pspec = g_param_spec_float("loss",
				   "The percentage of packets to be dropped.",
				   "Set/Get the percentage of packets to be dropped.",
				   0 /* minimum value */,
				   100 /* maximum value */,
				   0 /* default value */,
				   G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_NETEM_LOSS,
                                        pspec);

        pspec = g_param_spec_uint("gap",
				  "The reordering gap.",
				  "Set/Get the reordering gap.",
				  0 /* minimum value */,
				  G_MAXUINT /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_NETEM_GAP,
                                        pspec);

        pspec = g_param_spec_float("duplicate",
				   "The percentage of packets to randomly duplicate.",
				   "Set/Get the percentage of packets to randomly duplicate.",
				   0 /* minimum value */,
				   100 /* maximum value */,
				   0 /* default value */,
				   G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_NETEM_DUPLICATE,
                                        pspec);

        pspec = g_param_spec_uint("jitter",
				  "The amount of random jitter to add in ms.",
				  "Set/Get the amount of random jitter to add in ms.",
				  0 /* minimum value */,
				  G_MAXUINT /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_NETEM_JITTER,
                                        pspec);

        pspec = g_param_spec_float("latency-correlation",
				   "The amount that the previous packet effects the "
				   "next packets latency expressed as a precentage.",
				   "Set/Get the amount that the previous packet effects the "
				   "next packets latency expressed as a precentage.",
				   0 /* minimum value */,
				   100 /* maximum value */,
				   0 /* default value */,
				   G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_NETEM_LATENCY_CORR,
                                        pspec);

        pspec = g_param_spec_float("loss-correlation",
				   "The amount that the previous packet effects the "
				   "next packets loss probability expressed as a precentage.",
				   "Set/Get the amount that the previous packet effects the "
				   "next packets loss probability expressed as a percentage.",
				   0 /* minimum value */,
				   100 /* maximum value */,
				   0 /* default value */,
				   G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_NETEM_LOSS_CORR,
                                        pspec);

        pspec = g_param_spec_float("duplicate-correlation",
				   "The amount that the previous packet effects the "
				   "next packets duplication probability expressed as a "
				   "percentage.",
				   "Set/Get the amount that the previous packet effects the "
				   "next packets duplication probability expressed as a "
				   "percentage.",
				   0 /* minimum value */,
				   100 /* maximum value */,
				   0 /* default value */,
				   G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_NETEM_DUP_CORR,
                                        pspec);
}

GType lql_qdisc_netem_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLQDiscNetemClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_qdisc_netem_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLQDiscNetem),
                        0,      /* n_preallocs */
                        lql_qdisc_netem_instance_init    /* instance_init */
                };
                type = g_type_register_static (LQL_TYPE_QDISC,
                                               "LQLQDiscNetem",
                                               &info, 0);
        }
        return type;
}
