/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-qdisc-priomap.h"
#include "lql-qdisc-priomap-private.h"
#include "lql_qdisc_priomap_helper.h"

/********************************************************************
 * Non-Virutal API
 ********************************************************************/

/**
 * lql_qdisc_priomap_new:
 *
 * Build a new instance of LQLQDiscPriomap. 
 * 
 * Return value: A new LQLQDiscPriomap instance.
 **/
LQLQDiscPriomap*
lql_qdisc_priomap_new(void)
{
        return(LQL_QDISC_PRIOMAP(g_object_new(LQL_TYPE_QDISC_PRIOMAP, NULL)));
}

/**
 * lql_qdisc_priomap_get_bands:
 * @self: The LQLQDiscPriomap object to work with.
 * 
 * Get the number of traffic bands this LQLQDiscPriomap has.
 * 
 * Return value: The number of traffic bands.
 **/
int
lql_qdisc_priomap_get_bands(LQLQDiscPriomap *self)
{
        int bands = 0;

        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC_PRIOMAP(self), "bands", &bands, NULL);

        return(bands);
}

/**
 * lql_qdisc_priomap_set_bands:
 * @self: The LQLQDisc object to work with.
 * @bands: The number of bands.
 * 
 * Set the number of traffic bands for this LQLQDiscPriomap.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_priomap_set_bands(LQLQDiscPriomap *self,
			    int bands)
{
        g_assert(self != NULL);
	g_assert(bands > 0);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }
        
        g_object_set(LQL_QDISC_PRIOMAP(self), "bands", bands, NULL);
        
        return(TRUE);
}

/**
 * lql_qdisc_priomap_set_prio:
 * @self: The LQLQDiscPriomap instance.
 * @index: The index in the priority table to set.
 * @value: The value to set the priority table entry to.
 * 
 * This function is used to set the traffic band (class) that packets will be directed to
 * for a certain value (index) of the TOS bits.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_priomap_set_prio(LQLQDiscPriomap *self,
			   int index,
			   unsigned char value)
{
	g_assert(self != NULL);
	g_assert((index >= 0) && (index < PRIO_MAX_SIZE));

	self->private->priomap[index] = value;

	return(TRUE);
}

/**
 * lql_qdisc_priomap_get_prio:
 * @self: The LQLQDiscPriomap instance.
 * @index: The index in the priomap table to return.
 * 
 * This function returns the traffic band that packets with TOS value index will be
 * placed in.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
unsigned char
lql_qdisc_priomap_get_prio(LQLQDiscPriomap *self, 
			   int index)
{
	g_assert(self != NULL);
	g_assert((index >= 0) && (index < PRIO_MAX_SIZE));
		
	return(self->private->priomap[index]);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/**
 * lql_qdisc_priomap_fill_from_nlhdr:
 * @self: The LQLQDisc object to work with.
 * @hdr: The nl packet to get the queue information from.
 *
 * This method takes a nlmsg from the kernel and sets this
 * objects instance variables to match it. This is used
 * to create objects that match the ones currently active
 * in the system. This overrides the default LQLQDisc
 * implementation of this method.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_priomap_fill_from_nlhdr(LQLQDisc *self,
				  const struct nlmsghdr *hdr)
{
	g_assert(self != NULL);
	g_assert(hdr != NULL);

	return(lql_qdisc_priomap_helper_fill_from_nlhdr(LQL_QDISC_PRIOMAP(self), hdr));
}

/**
 * lql_qdisc_priomap_add:
 * @self: The LQLQDisc object to add to the system.
 * @con: The LQLCon connection to use when adding this queueing discipline.
 *
 * Add this PRIOMAP queueing discipline to the system. This overrides
 * the default implementation of this function in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_priomap_add(LQLQDisc *self,
		      LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_priomap_helper_add(LQL_QDISC_PRIOMAP(self), con));
}

/**
 * lql_qdisc_priomap_modify:
 * @self: The LQLQDisc object to modify.
 * @con: The LQLCon connection to use when adding this qdisc.
 *
 * Modify the PRIOMAP queueing discipline in the system to match the current
 * object state. This method overrides the default
 * implementation in LQLQDisc.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_priomap_modify(LQLQDisc *self,
			 LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);

	return(lql_qdisc_priomap_helper_modify(LQL_QDISC_PRIOMAP(self), con));
}

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_QDISC_PRIOMAP_BANDS = 1,
};

static void
lql_qdisc_priomap_instance_init(GTypeInstance *instance,
				gpointer g_class)
{
	LQLQDiscPriomap *self = (LQLQDiscPriomap *)instance;
	self->private = g_new0(LQLQDiscPriomapPrivate, 1);
	self->private->dispose_has_run = FALSE;

	/* Default number of bands is 3. */
	self->private->bands = 3;

	/* Set the default values for the priomap. These are the defaults
	   that the TC command uses. */
	self->private->priomap[0] = 1;
	self->private->priomap[1] = 2;
	self->private->priomap[2] = 2;
	self->private->priomap[3] = 2;
	self->private->priomap[4] = 1;
	self->private->priomap[5] = 2;
	self->private->priomap[6] = 0;
	self->private->priomap[7] = 0;
	self->private->priomap[8] = 1;
	self->private->priomap[9] = 1;
	self->private->priomap[10] = 1;
	self->private->priomap[11] = 1;
	self->private->priomap[12] = 1;
	self->private->priomap[13] = 1;
	self->private->priomap[14] = 1;
	self->private->priomap[15] = 1;
}

static void
lql_qdisc_priomap_set_property(GObject *object,
			       guint property_id,
			       const GValue *value,
			       GParamSpec *pspec)
{
        LQLQDiscPriomap *self = (LQLQDiscPriomap *) object;

        switch (property_id) {
        case LQL_QDISC_PRIOMAP_BANDS:
        {
                self->private->bands = g_value_get_int(value);
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static void
lql_qdisc_priomap_get_property(GObject *object,
			       guint property_id,
			       GValue *value,
			       GParamSpec *pspec)
{
        LQLQDiscPriomap *self = (LQLQDiscPriomap *) object;

        switch (property_id) {
        case LQL_QDISC_PRIOMAP_BANDS: 
        {
                g_value_set_int(value, self->private->bands);
                break;
        }
        default:
                g_assert(FALSE);
                break;
        }
}

static GObject *
lql_qdisc_priomap_constructor(GType type,
			      guint n_construct_properties,
			      GObjectConstructParam *construct_properties)
{
        GObject *obj = NULL;

        {
                /* Invoke parent constructor. */
		LQLQDiscPriomapClass *klass;
		GObjectClass *parent_class;  
		klass = LQL_QDISC_PRIOMAP_CLASS(g_type_class_peek(LQL_TYPE_QDISC_PRIOMAP));
		parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
		obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
lql_qdisc_priomap_dispose(GObject *obj)
{
        LQLQDiscPriomap *self = (LQLQDiscPriomap *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_qdisc_priomap_finalize(GObject *obj)
{
        LQLQDiscPriomap *self = (LQLQDiscPriomap *)obj;

        /*
         * Here, complete object destruction.
         */

        
        g_free (self->private);
}


static void
lql_qdisc_priomap_class_init(gpointer g_class,
			 gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        LQLQDiscClass *klass = LQL_QDISC_CLASS(g_class);
        GParamSpec *pspec = NULL;

        gobject_class->set_property = lql_qdisc_priomap_set_property;
        gobject_class->get_property = lql_qdisc_priomap_get_property;
        gobject_class->dispose = lql_qdisc_priomap_dispose;
        gobject_class->finalize = lql_qdisc_priomap_finalize;
        gobject_class->constructor = lql_qdisc_priomap_constructor;

	/* Set the virtual methods. */
	klass->fill_from_nlhdr = lql_qdisc_priomap_fill_from_nlhdr;
	klass->add = lql_qdisc_priomap_add;
	klass->modify = lql_qdisc_priomap_modify;
	klass->print = lql_qdisc_priomap_helper_print;

	/* Setup the properties. */
        pspec = g_param_spec_int("bands",
				 "The number of traffic bands this LQLQDiscPriomap has",
				 "Set/Get the number of traffic bands",
				 0 /* minimum value */,
				 G_MAXINT /* maximum value */,
				 3 /* default value */,
				 G_PARAM_READWRITE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_PRIOMAP_BANDS,
                                        pspec);
}

GType lql_qdisc_priomap_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLQDiscPriomapClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_qdisc_priomap_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLQDiscPriomap),
                        0,      /* n_preallocs */
                        lql_qdisc_priomap_instance_init    /* instance_init */
                };
                type = g_type_register_static (LQL_TYPE_QDISC,
                                               "LQLQDiscPriomap",
                                               &info, 0);
        }
        return type;
}
