/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "lql-qdisc.h"
#include "lql-qdisc-private.h"
#include "lql_qdisc_helper.h"

/********************************************************************
 * Non-Virtual API
 ********************************************************************/

/**
 * lql_qdisc_new:
 *
 * Build a new instance of LQLQDisc. 
 * 
 * Return value: A new LQLQDisc instance.
 **/
LQLQDisc*
lql_qdisc_new(void)
{
        return(LQL_QDISC(g_object_new(LQL_TYPE_QDISC, NULL)));
}

/**
 * lql_qdisc_set_parent:
 * @self: The LQLQDisc instance.
 * @parent: The LQLElement that is the parent of this LQLQDisc.
 * 
 * Sets the parent of this LQLQDisc to be the passed LQLElement.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_set_parent(LQLQDisc *self,
		     LQLElement *parent)
{
        g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	if (parent == NULL) {
		/* Assume setting parent to TC_H_ROOT. */
		g_object_set(LQL_ELEMENT(self), "parent-handle", TC_H_ROOT, NULL);
	} else {
		g_object_set(LQL_ELEMENT(self), "parent-handle",
			     lql_element_get_handle(LQL_ELEMENT(parent)), NULL);
	}
        
        return(TRUE);
}

/**
 * lql_qdisc_set_parent_class_id:
 * @self: 
 * @class: 
 * 
 * Set the class portion of the parent handle. This method is only useful in
 * situations where the underlying QDisc automatically creates classes that you
 * probably don't have a LQLClass instance for. Priomap in particular does this.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_set_parent_class_id(LQLQDisc *self,
			      guint16 class)
{
	guint32 tmp = 0;

	g_assert(self != NULL);
	
	g_object_get(LQL_ELEMENT(self), "parent-handle", &tmp, NULL);
	
	tmp = (tmp & 0xFFFF0000) | class;
	
	g_object_set(LQL_ELEMENT(self), "parent-handle", tmp, NULL);
	
	return(TRUE);
}

/**
 * lql_qdisc_set_id:
 * @self: The LQLQDisc instance.
 * @id: The ID of this LQLQDisc.
 * 
 * Set the ID number for this qdisc. In relation to TC this number is the part of the
 * QDisc identifier before the colon.
 * 
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
lql_qdisc_set_id(LQLQDisc *self,
		 guint16 id)
{
	guint32 tmp = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	g_object_get(LQL_ELEMENT(self), "handle", &tmp, NULL);
	
	tmp = (tmp & 0x0000FFFF) | (id << 16);
        
        g_object_set(LQL_ELEMENT(self), "handle", tmp, NULL);

	return(TRUE);
}


/**
 * lql_qdisc_get_id:
 * @self: The LQLQDisc instance.
 * 
 * Get the ID number of this QDisc. In relation to TC this number is the part of the
 * QDisc identifier before the colon.
 * 
 * Return value: The ID number of this QDisc.
 **/
guint16
lql_qdisc_get_id(LQLQDisc *self)
{
	guint32 handle = 0;
	guint16 id = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC(self), "handle", &handle, NULL);

	id = handle >> 16;

	return(id);
}

/**
 * lql_qdisc_get_bytes:
 * @self: The LQLQDisc instance.
 * 
 * Get the number of bytes that have been passed by this QDisc.
 * 
 * Return value: The number of bytes.
 **/
guint64
lql_qdisc_get_bytes(LQLQDisc *self)
{
	guint64 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC(self), "enqueued-bytes", &stat, NULL);

	return(stat);
}

/**
 * lql_qdisc_get_packets:
 * @self: The LQLQDisc instance.
 * 
 * Get the number of packets that have been passed by this QDisc.
 * 
 * Return value: The number of packets.
 **/
guint32
lql_qdisc_get_packets(LQLQDisc *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC(self), "enqueued-packets", &stat, NULL);

	return(stat);
}

/**
 * lql_qdisc_get_drops:
 * @self: The LQLQDisc instance.
 * 
 * Get the number of packets dropped by this QDisc.
 * 
 * Return value: The number of drops.
 **/
guint32
lql_qdisc_get_drops(LQLQDisc *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC(self), "dropped-packets", &stat, NULL);

	return(stat);
}

/**
 * lql_qdisc_get_overlimits:
 * @self: The LQLQDisc instance.
 * 
 * Get the number of over limits packets.
 * 
 * Return value: The number of over limits packets.
 **/
guint32
lql_qdisc_get_overlimits(LQLQDisc *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC(self), "overlimits", &stat, NULL);

	return(stat);
}

/**
 * lql_qdisc_get_bps:
 * @self: The LQLQDisc instance.
 * 
 * Get the current bytes per second rate for this QDisc.
 * 
 * Return value: The current BPS.
 **/
guint32
lql_qdisc_get_bps(LQLQDisc *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC(self), "bps", &stat, NULL);

	return(stat);
}

/**
 * lql_qdisc_get_pps:
 * @self: The LQLQDisc instance.
 * 
 * Get the current packets per second rate for this QDisc.
 * 
 * Return value: The current PPS.
 **/
guint32
lql_qdisc_get_pps(LQLQDisc *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC(self), "pps", &stat, NULL);

	return(stat);
}

/**
 * lql_qdisc_get_qlen:
 * @self: The LQLQDisc instance.
 * 
 * Get the current queue length for this QDisc.
 * 
 * Return value: The current queue length.
 **/
guint32
lql_qdisc_get_qlen(LQLQDisc *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC(self), "queue-length", &stat, NULL);

	return(stat);
}

/**
 * lql_qdisc_get_backlog:
 * @self: The LQLQDisc instance.
 * 
 * Get the current number of backlog packets.
 * 
 * Return value: The number of backlog packets.
 **/
guint32
lql_qdisc_get_backlog(LQLQDisc *self)
{
	guint32 stat = 0;

	g_assert(self != NULL);

        if (self->private->dispose_has_run) {
                return(0);
        }
        
        g_object_get(LQL_QDISC(self), "backlog", &stat, NULL);

	return(stat);
}

/********************************************************************
 * Virtual API
 ********************************************************************/

/**
 * lql_qdisc_fill_from_nlhdr:
 * @self: The LQLQDisc object to work with.
 * @hdr: The qdisc information from the kernel.
 * 
 * Fill in the instance variables for this instance of LQLQDisc
 * to match the network qdisc information in hdr. This function is used
 * by the LQLCon functions that get the list of QDiscs that are currently
 * configured on the system.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_fill_from_nlhdr(LQLQDisc *self,
			  const struct nlmsghdr *hdr)
{
        g_assert(self != NULL);
        g_assert(hdr != NULL);

        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_QDISC_GET_CLASS(self)->fill_from_nlhdr(self, hdr));
}

/**
 * lql_qdisc_add:
 * @self: The LQLQDisc object to add to the system.
 * @con: The LQLCon connection to use when adding this queueing discipline.
 *
 * Add this queueing discipline to the system.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_add(LQLQDisc *self,
	      LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_QDISC_GET_CLASS(self)->add(self, con));
}

/**
 * lql_qdisc_modify:
 * @self: The LQLQDisc object to modify.
 * @con: The LQLCon connection to use when adding this qdisc.
 *
 * Modify the queueing discipline in the system to match the current
 * object state.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_modify(LQLQDisc *self,
		 LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_QDISC_GET_CLASS(self)->modify(self, con));
}

/**
 * lql_qdisc_delete:
 * @self: The LQLQDisc object to delete from the system.
 * @con: The LQLCon connection to use when deleting this queueing discipline.
 *
 * Delete this queueing discipline from the system.
 *
 * Returns: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_delete(LQLQDisc *self,
		 LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_QDISC_GET_CLASS(self)->delete(self, con));
}

/**
 * lql_qdisc_print:
 * @self: The LQLQDisc instance.
 * 
 * Print out some information on this QDisc.
 **/
void
lql_qdisc_print(LQLQDisc *self)
{
	g_assert(self != NULL);
	
        if (self->private->dispose_has_run) {
                return;
        }

	LQL_QDISC_GET_CLASS(self)->print(self);
}

/**
 * lql_qdisc_printStats:
 * @self: The LQLQDisc instance.
 * 
 * Print out the statistics information for this QDisc.
 **/
void
lql_qdisc_printStats(LQLQDisc *self)
{
	g_assert(self != NULL);
	
        if (self->private->dispose_has_run) {
                return;
        }

	LQL_QDISC_GET_CLASS(self)->printStats(self);
}

/**
 * lql_qdisc_updateStats:
 * @self: The LQLQDisc instance.
 * @con: The LQLCon instance.
 * 
 * Update the statistics for this QDisc. This involves querying the information
 * from the kernel.
 * 
 * Return value: TRUE on success, FALSE on failure.
 **/
gboolean
lql_qdisc_updateStats(LQLQDisc *self,
		      LQLCon *con)
{
	g_assert(self != NULL);
	g_assert(con != NULL);
	
        if (self->private->dispose_has_run) {
                return(FALSE);
        }

	return(LQL_QDISC_GET_CLASS(self)->updateStats(self, con));
}

/********************************************************************
 * Object stuff.
 ********************************************************************/
enum {
        LQL_QDISC_BYTES = 1,
	LQL_QDISC_PACKETS = 2,
	LQL_QDISC_DROPS = 3,
	LQL_QDISC_OVERLIMITS = 4,
	LQL_QDISC_BPS = 5,
	LQL_QDISC_PPS = 6,
	LQL_QDISC_QLEN = 7,
	LQL_QDISC_BACKLOG = 8,
};

static void
lql_qdisc_instance_init(GTypeInstance *instance,
			gpointer g_class)
{
        LQLQDisc *self = (LQLQDisc *)instance;
        self->private = g_new0(LQLQDiscPrivate, 1);
        self->private->dispose_has_run = FALSE;
}

static void
lql_qdisc_set_property(GObject *object,
		       guint property_id,
		       const GValue *value,
		       GParamSpec *pspec)
{
	LQLQDisc *self = (LQLQDisc *) object;

        switch (property_id) {
	case LQL_QDISC_BYTES:
	{
                self->private->bytes = g_value_get_int64(value);		
		break;
	}
	case LQL_QDISC_PACKETS:
	{
		self->private->packets = g_value_get_uint(value);
		break;
	}
	case LQL_QDISC_DROPS:
	{
		self->private->drops = g_value_get_uint(value);
		break;
	}
	case LQL_QDISC_OVERLIMITS:
	{
		self->private->overlimits = g_value_get_uint(value);
		break;
	}
	case LQL_QDISC_BPS:
	{
		self->private->bps = g_value_get_uint(value);
		break;
	}
	case LQL_QDISC_PPS:
	{
		self->private->pps = g_value_get_uint(value);
		break;
	}
	case LQL_QDISC_QLEN:
	{
		self->private->qlen = g_value_get_uint(value);
		break;
	}
	case LQL_QDISC_BACKLOG:
	{
		self->private->backlog = g_value_get_uint(value);
		break;
	}
        default:
                g_assert(FALSE);
                break;
        }
}

static void
lql_qdisc_get_property(GObject *object,
		       guint property_id,
		       GValue *value,
		       GParamSpec *pspec)
{
        LQLQDisc *self = (LQLQDisc *) object;

        switch (property_id) {
	case LQL_QDISC_BYTES:
	{
                g_value_set_int64(value, self->private->bytes);
		break;
	}
	case LQL_QDISC_PACKETS:
	{
		g_value_set_uint(value, self->private->packets);
		break;
	}
	case LQL_QDISC_DROPS:
	{
		g_value_set_uint(value, self->private->drops);
		break;
	}
	case LQL_QDISC_OVERLIMITS:
	{
		g_value_set_uint(value, self->private->overlimits);
		break;
	}
	case LQL_QDISC_BPS:
	{
		g_value_set_uint(value, self->private->bps);
		break;
	}
	case LQL_QDISC_PPS:
	{
		g_value_set_uint(value, self->private->pps);
		break;
	}
	case LQL_QDISC_QLEN:
	{
		g_value_set_uint(value, self->private->qlen);
		break;
	}
	case LQL_QDISC_BACKLOG:
	{
		g_value_set_uint(value, self->private->backlog);
		break;
	}
        default:
                g_assert(FALSE);
                break;
        }
}

static GObject*
lql_qdisc_constructor(GType type,
		      guint n_construct_properties,
		      GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                LQLQDiscClass *klass;
                GObjectClass *parent_class;  
                klass = LQL_QDISC_CLASS(g_type_class_peek(LQL_TYPE_QDISC));
                parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
                obj = parent_class->constructor(type,
						n_construct_properties,
						construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
lql_qdisc_dispose (GObject *obj)
{
        LQLQDisc *self = (LQLQDisc *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
lql_qdisc_finalize (GObject *obj)
{
        LQLQDisc *self = (LQLQDisc *)obj;

        /*
         * Here, complete object destruction.
         * You might not need to do much...
         */
        
        g_free (self->private);
}

static void
lql_qdisc_class_init(gpointer g_class,
		     gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
        LQLQDiscClass *klass = LQL_QDISC_CLASS(g_class);
        GParamSpec *pspec = NULL;

        gobject_class->set_property = lql_qdisc_set_property;
        gobject_class->get_property = lql_qdisc_get_property;
        gobject_class->dispose = lql_qdisc_dispose;
        gobject_class->finalize = lql_qdisc_finalize;
        gobject_class->constructor = lql_qdisc_constructor;

	/* Set the virtual methods. */
	klass->fill_from_nlhdr = NULL;
	klass->add = NULL;
	klass->modify = NULL;
	klass->delete = lql_qdisc_helper_delete;
	klass->print = lql_qdisc_helper_print;
	klass->printStats = lql_qdisc_helper_printStats;
	klass->updateStats = lql_qdisc_helper_updateStats;

	/* Setup the properties. */
        pspec = g_param_spec_int64("enqueued-bytes",
				   "The count of how many bytes have been enqueued",
				   "Get the cound of how many bytes have been enqueued.",
				   G_MININT64 /* minimum value */,
				   G_MAXINT64 /* maximum value */,
				   0 /* default value */,
				   G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_BYTES,
                                        pspec);

        pspec = g_param_spec_uint("enqueued-packets",
				  "The count of how many packets have been enqueued",
				  "Get the count of how many packets have been enqueued.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_PACKETS,
                                        pspec);

        pspec = g_param_spec_uint("dropped-packets",
				  "The count of how many packets have been dropped",
				  "Get the count of how many packets have been dropped.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_DROPS,
                                        pspec);

        pspec = g_param_spec_uint("overlimits",
				  "The count of how many packets were overlimits",
				  "Get the count of how many packets were overlimits.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_OVERLIMITS,
                                        pspec);

        pspec = g_param_spec_uint("bps",
				  "Rate in bytes per second",
				  "Get the rate in bytes per second.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_BPS,
                                        pspec);

        pspec = g_param_spec_uint("pps",
				  "The rate in packets per second",
				  "Get the rate in packets per second.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_PPS,
                                        pspec);

        pspec = g_param_spec_uint("queue-length",
				  "The length of the packet queue.",
				  "Get the length of the packet queue.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_QLEN,
                                        pspec);

        pspec = g_param_spec_uint("backlog",
				  "The number of backlog packets",
				  "Get the number of backlog packets.",
				  0 /* minimum value */,
				  G_MAXUINT32 /* maximum value */,
				  0 /* default value */,
				  G_PARAM_READABLE);
        g_object_class_install_property(gobject_class,
                                        LQL_QDISC_BACKLOG,
                                        pspec);
}

GType lql_qdisc_get_type(void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (LQLQDiscClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        lql_qdisc_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (LQLQDisc),
                        0,      /* n_preallocs */
                        lql_qdisc_instance_init    /* instance_init */
                };
                type = g_type_register_static(LQL_TYPE_ELEMENT,
					      "LQLQDisc",
					      &info, 0);
        }

        return type;
}
