/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <unistd.h>
#include <string.h>

#include "debug_macros.h"
#include "nl.h"
#include "lql-con.h"
#include "lql-interface.h"
#include "lql-class-dsmark.h"
#include "lql-class-dsmark-private.h"

gboolean
lql_class_dsmark_helper_fill_from_nlhdr(LQLClassDSMark *dsmark,
					const struct nlmsghdr *hdr)
{
	guint8 *tmp = NULL;

	g_assert(dsmark != NULL);
	g_assert(hdr != NULL);

	lql_element_set_handle(LQL_ELEMENT(dsmark), nl_tcmsg_get_handle(hdr));
	lql_element_set_parentHandle(LQL_ELEMENT(dsmark), nl_tcmsg_get_parent(hdr));
	lql_element_set_ifIndex(LQL_ELEMENT(dsmark), nl_tcmsg_get_ifindex(hdr));

	tmp = (guint8 *) nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_DSMARK_MASK);
	if (tmp == NULL) {
		g_printerr("Could not get TCA_DSMARK_MASK\n");
		return(FALSE);
	}
	lql_class_dsmark_set_mask(dsmark, *tmp);

	tmp = (guint8 *) nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_DSMARK_VALUE);
	if (tmp == NULL) {
		g_printerr("Could not get TCA_DSMARK_VALUE\n");
		return(FALSE);
	}
	lql_class_dsmark_set_value(dsmark, *tmp);

	return(TRUE);
}

static gboolean
lql_class_dsmark_helper_do(LQLClassDSMark *dsmark,
			   LQLCon *con,
			   int flags)
{
	TcPacket *pkt = NULL;
	struct nlmsghdr *hdr = NULL;
	gchar *kind = "dsmark";
	guint8 tmp = 0;

	g_assert(con != NULL);
	g_assert(dsmark != NULL);

	pkt = nl_tcpacket_new();
	hdr = nl_tcpacket_get_nlmsg(pkt);
	
	nl_nlmsg_set_type(hdr, RTM_NEWTCLASS);
	nl_nlmsg_set_flags(hdr, NLM_F_REQUEST | NLM_F_ACK | flags);
	nl_nlmsg_set_pid(hdr, getpid());
	nl_nlmsg_set_seq(hdr, lql_con_get_inc_sequence(LQL_CON(con)));
	
	nl_tcmsg_set_family(hdr, AF_UNSPEC);
	nl_tcmsg_set_ifindex(hdr, lql_element_get_ifIndex(LQL_ELEMENT(dsmark)));
	nl_tcmsg_set_parent(hdr, lql_element_get_parentHandle(LQL_ELEMENT(dsmark)));
	nl_tcmsg_set_handle(hdr, lql_element_get_handle(LQL_ELEMENT(dsmark)));

	nl_tcpacket_add_rtattr(pkt, TCA_KIND, strlen(kind) + 1, kind);
	nl_tcpacket_add_rtattr(pkt, TCA_OPTIONS, 0, NULL);
	tmp = lql_class_dsmark_get_mask(dsmark);
	nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_DSMARK_MASK,
				   NLMSG_ALIGN(sizeof(guint8)), &tmp);
	tmp = lql_class_dsmark_get_value(dsmark);
	nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_DSMARK_VALUE,
				   NLMSG_ALIGN(sizeof(guint8)), &tmp);

	if (!nl_tcpacket_do_command(pkt, lql_con_get_nlcon(LQL_CON(con)))) {
		return(FALSE);
	}

	g_free(pkt);

	return(TRUE);
}

gboolean
lql_class_dsmark_helper_add(LQLClassDSMark *class,
			    LQLCon *con)
{
	/* Since DSMark classes cannot be added, they always exist, This method
	   has the same flags as modify. */
	return(lql_class_dsmark_helper_do(class, con, 0));
}

gboolean
lql_class_dsmark_helper_modify(LQLClassDSMark *class,
			    LQLCon *con)
{
	return(lql_class_dsmark_helper_do(class, con, 0));
}

void
lql_class_dsmark_helper_print(LQLClass *self)
{
	LQLClassDSMarkClass *klass = NULL;
	LQLClassClass *parentKlass = NULL;
	LQLClassDSMark *dsmark = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_CLASS_DSMARK(self));

	dsmark = LQL_CLASS_DSMARK(self);

        if (dsmark->private->dispose_has_run) {
                return;
        }

	klass = LQL_CLASS_DSMARK_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	parentKlass->print(self);

	g_print("Type: DSMark\n");
	g_print("DSMark mask: %u\n", lql_class_dsmark_get_mask(dsmark));
	g_print("DSMark value: %u\n", lql_class_dsmark_get_value(dsmark));
}

void
lql_class_dsmark_helper_printStats(LQLClass *self)
{
	LQLClassDSMarkClass *klass = NULL;
	LQLClassClass *parentKlass = NULL;
	LQLClassDSMark *dsmark = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_CLASS_DSMARK(self));

	dsmark = LQL_CLASS_DSMARK(self);

        if (dsmark->private->dispose_has_run) {
                return;
        }

	klass = LQL_CLASS_DSMARK_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	parentKlass->printStats(self);
}

gboolean
lql_class_dsmark_helper_updateStats(LQLClass *self,
				    LQLCon *con)
{
	LQLClassDSMarkClass *klass = NULL;
	LQLClassClass *parentKlass = NULL;
	LQLClassDSMark *dsmark = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_CLASS_DSMARK(self));

	dsmark = LQL_CLASS_DSMARK(self);

        if (dsmark->private->dispose_has_run) {
                return(FALSE);
        }

	klass = LQL_CLASS_DSMARK_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	return(parentKlass->updateStats(self, con));
}
