/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <unistd.h>
#include <string.h>

#include "debug_macros.h"
#include "nl.h"
#include "lql-con.h"
#include "lql-interface.h"
#include "lql-class-htb.h"
#include "lql-class-htb-private.h"

#define HTB_TC_VER 0x30000
#if HTB_TC_VER >> 16 != TC_HTB_PROTOVER
#error "Different kernel and TC HTB versions"
#endif

gboolean
lql_class_htb_helper_fill_from_nlhdr(LQLClassHTB *htb,
				     const struct nlmsghdr *hdr)
{
	struct tc_htb_opt *opt = NULL;

	g_assert(htb != NULL);
	g_assert(hdr != NULL);

	lql_element_set_handle(LQL_ELEMENT(htb), nl_tcmsg_get_handle(hdr));
	lql_element_set_parentHandle(LQL_ELEMENT(htb), nl_tcmsg_get_parent(hdr));
	lql_element_set_ifIndex(LQL_ELEMENT(htb), nl_tcmsg_get_ifindex(hdr));

	opt = (struct tc_htb_opt *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_HTB_PARMS);
	if (opt == NULL) {
		g_printerr("Could not get HTB class parameters.\n");
		return(FALSE);
	}

	lql_class_htb_set_rate(htb, opt->rate.rate);
	lql_class_htb_set_ceiling(htb, opt->ceil.rate);
	lql_class_set_priority(LQL_CLASS(htb), opt->prio);

	return(TRUE);
}

static gboolean
lql_class_htb_helper_do(LQLClassHTB *htb,
			LQLCon *con,
			int flags)
{
	TcPacket *pkt = NULL;
	struct nlmsghdr *hdr = NULL;
	gchar *kind = "htb";
	struct tc_htb_opt *opt;
	__u32 rtab[256], ctab[256];
	int mtu = 0;
	unsigned short mpu = 0;
	unsigned buffer = 0, cbuffer = 0;

	g_assert(con != NULL);
	g_assert(htb != NULL);

	mtu = lql_class_htb_get_mtu(LQL_CLASS_HTB(htb));
	mpu = lql_class_htb_get_mpu(LQL_CLASS_HTB(htb));

	opt = g_malloc0(sizeof(struct tc_htb_opt));

	opt->rate.rate = lql_class_htb_get_rate(htb);
	opt->rate.mpu = mpu;
	opt->ceil.rate = lql_class_htb_get_ceiling(htb);
	if (opt->ceil.rate == 0) {
		opt->ceil.rate = opt->rate.rate;
	}
	opt->ceil.mpu = mpu;
	opt->rate.cell_log = lql_con_get_cell_log(con, mtu);
	opt->ceil.cell_log = lql_con_get_cell_log(con, mtu);
	opt->prio = lql_class_get_priority(LQL_CLASS(htb));
	buffer = (opt->rate.rate / lql_con_get_hz(con)) + mtu;
	opt->buffer = lql_con_calc_xmit_time(con, opt->rate.rate, buffer);
	cbuffer = (opt->ceil.rate / lql_con_get_hz(con)) + mtu;
	opt->cbuffer = lql_con_calc_xmit_time(con, opt->ceil.rate, cbuffer);

	lql_con_calc_rate_table(con, opt->rate.rate, rtab, mtu, mpu);
	lql_con_calc_rate_table(con, opt->ceil.rate, ctab, mtu, mpu);
	
	pkt = nl_tcpacket_new();
	hdr = nl_tcpacket_get_nlmsg(pkt);
	
	nl_nlmsg_set_type(hdr, RTM_NEWTCLASS);
	nl_nlmsg_set_flags(hdr, NLM_F_REQUEST | NLM_F_ACK | flags);
	nl_nlmsg_set_pid(hdr, getpid());
	nl_nlmsg_set_seq(hdr, lql_con_get_inc_sequence(LQL_CON(con)));
	
	nl_tcmsg_set_family(hdr, AF_UNSPEC);
	nl_tcmsg_set_ifindex(hdr, lql_element_get_ifIndex(LQL_ELEMENT(htb)));
	nl_tcmsg_set_parent(hdr, lql_element_get_parentHandle(LQL_ELEMENT(htb)));
	nl_tcmsg_set_handle(hdr, lql_element_get_handle(LQL_ELEMENT(htb)));

	nl_tcpacket_add_rtattr(pkt, TCA_KIND, strlen(kind) + 1, kind);
	nl_tcpacket_add_rtattr(pkt, TCA_OPTIONS, 0, NULL);
	nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_HTB_PARMS,
				   NLMSG_ALIGN(sizeof(struct tc_htb_opt)), opt);
	nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_HTB_RTAB, 1024, rtab);
	nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_HTB_CTAB, 1024, ctab);

	if (!nl_tcpacket_do_command(pkt, lql_con_get_nlcon(LQL_CON(con)))) {
		return(FALSE);
	}

	g_free(pkt);
	g_free(opt);

	return(TRUE);
}

gboolean
lql_class_htb_helper_add(LQLClassHTB *class,
			 LQLCon *con)
{
	return(lql_class_htb_helper_do(class, con, NLM_F_CREATE | NLM_F_EXCL));
}

gboolean
lql_class_htb_helper_modify(LQLClassHTB *class,
			    LQLCon *con)
{
	return(lql_class_htb_helper_do(class, con, 0));
}

void
lql_class_htb_helper_print(LQLClass *self)
{
	LQLClassHTBClass *klass = NULL;
	LQLClassClass *parentKlass = NULL;
	LQLClassHTB *htb = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_CLASS_HTB(self));

	htb = LQL_CLASS_HTB(self);

        if (htb->private->dispose_has_run) {
                return;
        }

	klass = LQL_CLASS_HTB_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	parentKlass->print(self);

	g_print("Type: HTB\n");
	g_print("HTB rate: %u\n", lql_class_htb_get_rate(htb));
	g_print("HTB ceiling: %u\n", lql_class_htb_get_ceiling(htb));
}

void
lql_class_htb_helper_printStats(LQLClass *self)
{
	LQLClassHTBClass *klass = NULL;
	LQLClassClass *parentKlass = NULL;
	LQLClassHTB *htb = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_CLASS_HTB(self));

	htb = LQL_CLASS_HTB(self);

        if (htb->private->dispose_has_run) {
                return;
        }

	klass = LQL_CLASS_HTB_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	parentKlass->printStats(self);
}

gboolean
lql_class_htb_helper_updateStats(LQLClass *self,
				 LQLCon *con)
{
	LQLClassHTBClass *klass = NULL;
	LQLClassClass *parentKlass = NULL;
	LQLClassHTB *htb = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_CLASS_HTB(self));

	htb = LQL_CLASS_HTB(self);

        if (htb->private->dispose_has_run) {
                return(FALSE);
        }

	klass = LQL_CLASS_HTB_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	return(parentKlass->updateStats(self, con));
}
