/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <unistd.h>
#include <string.h>

#include "debug_macros.h"
#include "nl.h"
#include "lql-con.h"
#include "lql-interface.h"
#include "lql-classifier-fwmark.h"
#include "lql-classifier-fwmark-private.h"

gboolean
lql_classifier_fwmark_helper_set_mark(LQLClassifierFwmark *self,
				      guint32 mark)
{
	g_assert(self != NULL);

	lql_classifier_set_id(LQL_CLASSIFIER(self), mark);
	
	return(TRUE);
}

guint32
lql_classifier_fwmark_helper_get_mark(LQLClassifierFwmark *self)
{
	g_assert(self != NULL);
	
	return(lql_classifier_get_id(LQL_CLASSIFIER(self)));
}

gboolean
lql_classifier_fwmark_helper_fill_from_nlhdr(LQLClassifierFwmark *self,
					     const struct nlmsghdr *hdr)
{
	__u32 *class = NULL;

	g_assert(self != NULL);
	g_assert(hdr != NULL);

	lql_classifier_set_ifIndex(LQL_CLASSIFIER(self), nl_tcmsg_get_ifindex(hdr));
	lql_classifier_set_parentHandle(LQL_CLASSIFIER(self), nl_tcmsg_get_parent(hdr));
	lql_classifier_set_id(LQL_CLASSIFIER(self), nl_tcmsg_get_handle(hdr));
	lql_classifier_set_priority(LQL_CLASSIFIER(self), nl_tcmsg_get_info(hdr) >> 16);
	lql_classifier_set_protocol(LQL_CLASSIFIER(self),
				    (nl_tcmsg_get_info(hdr) & 0x0000FFFF));

	class = (__u32 *) nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_FW_CLASSID);
	if (class != NULL) {
		lql_classifier_set_classHandle(LQL_CLASSIFIER(self), *class);
	}

	return(TRUE);
}

static gboolean
lql_classifier_fwmark_helper_do(LQLClassifierFwmark *fwmark,
				LQLCon *con,
				int flags)
{
	TcPacket *pkt = NULL;
	struct nlmsghdr *nlmsg = NULL;
	char *kind = "fw";
	__u32 class = 0;
	
	g_assert(con != NULL);
	g_assert(fwmark != NULL);

	pkt = nl_tcpacket_new();
	nlmsg = nl_tcpacket_get_nlmsg(pkt);

	nl_nlmsg_set_type(nlmsg, RTM_NEWTFILTER);
	nl_nlmsg_set_flags(nlmsg, NLM_F_REQUEST | NLM_F_ACK | flags);
	nl_nlmsg_set_pid(nlmsg, getpid());
	nl_nlmsg_set_seq(nlmsg, lql_con_get_inc_sequence(LQL_CON(con)));

	nl_tcmsg_set_family(nlmsg, AF_UNSPEC);
	nl_tcmsg_set_ifindex(nlmsg, lql_classifier_get_ifIndex(LQL_CLASSIFIER(fwmark)));
	nl_tcmsg_set_parent(nlmsg, lql_classifier_get_parentHandle(LQL_CLASSIFIER(fwmark)));
	nl_tcmsg_set_handle(nlmsg, lql_classifier_get_id(LQL_CLASSIFIER(fwmark)));
	nl_tcmsg_set_info(nlmsg,
			  TC_H_MAKE(lql_classifier_get_priority(LQL_CLASSIFIER(fwmark))<<16,
				    lql_classifier_get_protocol(LQL_CLASSIFIER(fwmark))));

	if (!nl_tcpacket_add_rtattr(pkt, TCA_KIND, strlen(kind) + 1, kind)) {
		#if (DEBUG_LEVEL > 5)
		g_print("lql_classifier_fwmark_helper_do: Could not add"
			" TCA_KIND rtattr\n");
		#endif
		g_free(pkt);
		return(FALSE);
	}
	if (!nl_tcpacket_add_rtattr(pkt, TCA_OPTIONS, 0, NULL)) {
		#if (DEBUG_LEVEL > 5)
		g_print("lql_classifier_fwmark_helper_do: Could not add"
			" TCA_OPTIONS rtattr\n");
		#endif
		g_free(pkt);
		return(FALSE);
	}
	class = lql_classifier_get_classHandle(LQL_CLASSIFIER(fwmark));
	if (class != 0) {
		if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_FW_CLASSID,
						sizeof(class), &class)) {
#if (DEBUG_LEVEL > 5)
			g_print("lql_classifier_fwmark_helper_do: Could not add"
				" TCA_FW_CLASSID rtattr\n");
#endif
			g_free(pkt);
			return(FALSE);
		}
		if (!nl_tcpacket_do_command(pkt, lql_con_get_nlcon(LQL_CON(con)))) {
#if (DEBUG_LEVEL > 5)
			g_print("lql_classifier_fwmark_helper_do: Could not do command.\n");
#endif
			g_free(pkt);
			return(FALSE);
		}
	}

	g_free(pkt);
	
	return(TRUE);
}

gboolean
lql_classifier_fwmark_helper_add(LQLClassifierFwmark *fwmark,
				 LQLCon *con)
{
	g_assert(fwmark != NULL);
	g_assert(con != NULL);

	return(lql_classifier_fwmark_helper_do(fwmark, con, NLM_F_CREATE | NLM_F_EXCL));
}

gboolean
lql_classifier_fwmark_helper_modify(LQLClassifierFwmark *fwmark,
				    LQLCon *con)
{
	g_assert(fwmark != NULL);
	g_assert(con != NULL);

	return(lql_classifier_fwmark_helper_do(fwmark, con, 0));
}

gboolean
lql_classifier_fwmark_helper_delete(LQLClassifierFwmark *fwmark,
				    LQLCon *con)
{
	g_printerr("Cannot delete fwmarks yet.\n");
	return(FALSE);
}
