/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <unistd.h>
#include <string.h>

#include "debug_macros.h"
#include "nl.h"
#include "lql-con.h"
#include "lql-interface.h"
#include "lql-classifier-tcindex.h"
#include "lql-classifier-tcindex-private.h"

gboolean
lql_classifier_tcindex_helper_set_mark(LQLClassifierTCIndex *self,
				       guint32 mark)
{
	g_assert(self != NULL);

	lql_classifier_set_id(LQL_CLASSIFIER(self), mark);
	
	return(TRUE);
}

guint32
lql_classifier_tcindex_helper_get_mark(LQLClassifierTCIndex *self)
{
	g_assert(self != NULL);
	
	return(lql_classifier_get_id(LQL_CLASSIFIER(self)));
}

gboolean
lql_classifier_tcindex_helper_fill_from_nlhdr(LQLClassifierTCIndex *self,
					      const struct nlmsghdr *hdr)
{
	guint32 *tmp = 0;
	guint16 *tmp2 = 0;

	g_assert(self != NULL);
	g_assert(hdr != NULL);

	lql_classifier_set_ifIndex(LQL_CLASSIFIER(self), nl_tcmsg_get_ifindex(hdr));
	lql_classifier_set_parentHandle(LQL_CLASSIFIER(self), nl_tcmsg_get_parent(hdr));
	lql_classifier_set_id(LQL_CLASSIFIER(self), nl_tcmsg_get_handle(hdr));
	lql_classifier_set_priority(LQL_CLASSIFIER(self), nl_tcmsg_get_info(hdr) >> 16);
	lql_classifier_set_protocol(LQL_CLASSIFIER(self),
				    (nl_tcmsg_get_info(hdr) & 0x0000FFFF));

	tmp = (guint32 *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_TCINDEX_HASH);
	if (tmp != NULL) {
		lql_classifier_tcindex_set_hash(self, *tmp);
	}

	tmp2 = (guint16 *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_TCINDEX_MASK);
	if (tmp2 != NULL) {
		lql_classifier_tcindex_set_mask(self, *tmp2);
	}

	tmp = (guint32 *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_TCINDEX_SHIFT);
	if (tmp != NULL) {
		lql_classifier_tcindex_set_shift(self, *tmp);
	}

	tmp = (guint32 *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_TCINDEX_FALL_THROUGH);
	if (tmp != NULL) {
		lql_classifier_tcindex_set_fallThrough(self, TRUE);
	}

	tmp = (guint32 *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_TCINDEX_CLASSID);
	if (tmp != NULL) {
		/* I don't want to expose public API for this so just use the struct. */
		self->private->class = *tmp;
	}

	return(TRUE);
}

static gboolean
lql_classifier_tcindex_helper_do(LQLClassifierTCIndex *tcindex,
				 LQLCon *con,
				 int flags)
{
	TcPacket *pkt = NULL;
	struct nlmsghdr *nlmsg = NULL;
	char *kind = "tcindex";
	guint32 tmp = 0;
	guint16 tmp2 = 0;
	
	g_assert(con != NULL);
	g_assert(tcindex != NULL);

	pkt = nl_tcpacket_new();
	nlmsg = nl_tcpacket_get_nlmsg(pkt);

	nl_nlmsg_set_type(nlmsg, RTM_NEWTFILTER);
	nl_nlmsg_set_flags(nlmsg, NLM_F_REQUEST | NLM_F_ACK | flags);
	nl_nlmsg_set_pid(nlmsg, getpid());
	nl_nlmsg_set_seq(nlmsg, lql_con_get_inc_sequence(LQL_CON(con)));

	nl_tcmsg_set_family(nlmsg, AF_UNSPEC);
	nl_tcmsg_set_ifindex(nlmsg, lql_classifier_get_ifIndex(LQL_CLASSIFIER(tcindex)));
	nl_tcmsg_set_parent(nlmsg, lql_classifier_get_parentHandle(LQL_CLASSIFIER(tcindex)));
	nl_tcmsg_set_handle(nlmsg, lql_classifier_get_id(LQL_CLASSIFIER(tcindex)));
	nl_tcmsg_set_info(nlmsg,
			  TC_H_MAKE(lql_classifier_get_priority(LQL_CLASSIFIER(tcindex))<<16,
				    lql_classifier_get_protocol(LQL_CLASSIFIER(tcindex))));

	if (!nl_tcpacket_add_rtattr(pkt, TCA_KIND, strlen(kind) + 1, kind)) {
		g_print("lql_classifier_tcindex_helper_do: Could not add"
			" TCA_KIND rtattr\n");
		g_free(pkt);
		return(FALSE);
	}

	if (!nl_tcpacket_add_rtattr(pkt, TCA_OPTIONS, 0, NULL)) {
		g_print("lql_classifier_tcindex_helper_do: Could not add"
			" TCA_OPTIONS rtattr\n");
		g_free(pkt);
		return(FALSE);
	}

	tmp = lql_classifier_tcindex_get_hash(tcindex);
	if (tmp != 0) {
		if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_TCINDEX_HASH,
						sizeof(tmp), &tmp)) {
			g_print("lql_classifier_tcindex_helper_do: Could not add"
				" TCA_INDEX_HASH rtattr\n");
			g_free(pkt);
			return(FALSE);
		}
	}

	tmp2 = lql_classifier_tcindex_get_mask(tcindex);
	if (tmp2 != 0) {
		if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_TCINDEX_MASK,
						sizeof(tmp2), &tmp2)) {
			g_print("lql_classifier_tcindex_helper_do: Could not add"
				" TCA_INDEX_MASK rtattr\n");
			g_free(pkt);
			return(FALSE);
		}
	}

	tmp = lql_classifier_tcindex_get_shift(tcindex);
	if (tmp != 0) {
		if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_TCINDEX_SHIFT,
						sizeof(tmp), &tmp)) {
			g_print("lql_classifier_tcindex_helper_do: Could not add"
				" TCA_INDEX_SHIFT rtattr\n");
			g_free(pkt);
			return(FALSE);
		}
	}

	if (lql_classifier_tcindex_get_fallThrough(tcindex)) {
		/* It's on. */
		tmp = 1;
		if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_TCINDEX_FALL_THROUGH,
						sizeof(tmp), &tmp)) {
			g_print("lql_classifier_tcindex_helper_do: Could not add"
				" TCA_INDEX_FALL_THROUGH rtattr\n");
			g_free(pkt);
			return(FALSE);
		}
	} else {
		/* Not enabled. */
		tmp = 0;
		if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_TCINDEX_FALL_THROUGH,
						sizeof(tmp), &tmp)) {
			g_print("lql_classifier_tcindex_helper_do: Could not add"
				" TCA_INDEX_FALL_THROUGH rtattr\n");
			g_free(pkt);
			return(FALSE);
		}
	}

	tmp = lql_classifier_tcindex_get_classHandle(tcindex);
	if (tmp != 0) {
		if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_TCINDEX_CLASSID,
						sizeof(tmp), &tmp)) {
			g_print("lql_classifier_tcindex_helper_do: Could not add"
				" TCA_INDEX_CLASS rtattr\n");

			g_free(pkt);
			return(FALSE);
		}
	}

	if (!nl_tcpacket_do_command(pkt, lql_con_get_nlcon(LQL_CON(con)))) {
		g_printerr("lql_classifier_tcindex_helper_do: Could not do command.\n");
		g_free(pkt);
		return(FALSE);
	}
	
	g_free(pkt);
	
	return(TRUE);
}

gboolean
lql_classifier_tcindex_helper_add(LQLClassifierTCIndex *tcindex,
				  LQLCon *con)
{
	g_assert(tcindex != NULL);
	g_assert(con != NULL);

	return(lql_classifier_tcindex_helper_do(tcindex, con, NLM_F_CREATE | NLM_F_EXCL));
}

gboolean
lql_classifier_tcindex_helper_modify(LQLClassifierTCIndex *tcindex,
				     LQLCon *con)
{
	g_assert(tcindex != NULL);
	g_assert(con != NULL);

	return(lql_classifier_tcindex_helper_do(tcindex, con, 0));
}

gboolean
lql_classifier_tcindex_helper_delete(LQLClassifierTCIndex *tcindex,
				     LQLCon *con)
{
	g_assert(0);
	return(FALSE);
}
