/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <unistd.h>
#include <string.h>
#include <arpa/inet.h>
#include <glib.h>

#include "nl.h"
#include "lql-con.h"
#include "lql-classifier-u32.h"
#include "lql-classifier-u32-private.h"
#include "lql-u32match.h"

#define U32_BUFF 4096

#define IP_DST_OFFSET 16

/**
 * Type used to hold the information necessary to pass to the kernel to add a U32 classifier.
 */
typedef struct
{
	struct tc_u32_sel sel;
	struct tc_u32_key keys[];
} U32Sel;

gboolean
lql_classifier_u32_helper_set_handle(LQLClassifierU32 *classifier,
				     guint16 handle)
{
	guint32 tmp = 0;

	g_assert(classifier != NULL);

	tmp = handle;
	lql_classifier_set_id(LQL_CLASSIFIER(classifier),
			      ((handle << 20) |
			       (lql_classifier_get_id(LQL_CLASSIFIER(classifier)) &
				0x000FFFFF)));

	return(TRUE);
}

guint16
lql_classifier_u32_helper_get_handle(LQLClassifierU32 *classifier)
{
	g_assert(classifier != NULL);
	
	return(lql_classifier_get_id(LQL_CLASSIFIER(classifier)) >> 20);
}

gboolean
lql_classifier_u32_helper_set_hashHandle(LQLClassifierU32 *classifier,
					 guint16 hashHandle)
{
	g_assert(classifier != NULL);
	
	lql_classifier_set_id(LQL_CLASSIFIER(classifier),
			      (lql_classifier_get_id(LQL_CLASSIFIER(classifier)) & 0xFFF00FFF)
			      | (hashHandle << 12));
	
	return(TRUE);
}

guint16
lql_classifier_u32_helper_get_hashHandle(LQLClassifierU32 *classifier)
{
	g_assert(classifier != NULL);

	return((lql_classifier_get_id(LQL_CLASSIFIER(classifier)) & 0x000FF000) >> 12);
}

gboolean
lql_classifier_u32_helper_set_hashNode(LQLClassifierU32 *classifier,
				       guint16 hashNode)
{
	g_assert(classifier != NULL);
	
	lql_classifier_set_id(LQL_CLASSIFIER(classifier),
			      (lql_classifier_get_id(LQL_CLASSIFIER(classifier)) & 0xFFFFF000)
			      | hashNode);
	
	return(TRUE);
}

guint16
lql_classifier_u32_helper_get_hashNode(LQLClassifierU32 *classifier)
{
	g_assert(classifier != NULL);

	return(lql_classifier_get_id(LQL_CLASSIFIER(classifier)) & 0x00000FFF);
}

static gboolean
lql_classifier_u32_helper_add_keys(LQLClassifierU32 *u32,
				   U32Sel *sel,
				   unsigned int bufLen)
{
	int i;
	LQLU32Match *match = NULL;

	g_assert(u32 != NULL);
	g_assert(sel != NULL);

	for (i = 0; i < sel->sel.nkeys; i++) {
		match = lql_classifier_u32_get_match(u32, i);
		if (match == NULL) {
			// This shouldn't be needed.
			break;
		}

		sel->keys[i].val = lql_u32match_get_value(match);
		sel->keys[i].mask = lql_u32match_get_mask(match);
		sel->keys[i].off = lql_u32match_get_offset(match);
	}

	return(TRUE);
}

static U32Sel*
lql_classifier_u32_helper_build_sel(LQLClassifierU32 *u32,
				    int *len)
{
	U32Sel *sel = NULL;
	
	g_assert(u32 != NULL);

	sel = g_malloc0(U32_BUFF);

	if ((lql_classifier_u32_get_hashOffset(u32)) ||
	     (lql_classifier_u32_get_hashMask(u32))) {
		sel->sel.hoff = lql_classifier_u32_get_hashOffset(u32);
		sel->sel.hmask = htonl(lql_classifier_u32_get_hashMask(u32));
	} else {
		sel->sel.flags = TC_U32_TERMINAL;
	}
	     
	sel->sel.nkeys = lql_classifier_u32_get_match_len(LQL_CLASSIFIER_U32(u32));
	if (sel->sel.nkeys > 0) {
		// Make sure the buffer is big enough.
		*len = ((sel->sel.nkeys) * sizeof(struct tc_u32_key)) +
			sizeof(struct tc_u32_key);
		if (*len > U32_BUFF) {
			g_printerr("lql_con_helper_u32_build_sel: Too many matches.\n");
			return(NULL);
		}

		// Now add the keys to the buffer.
		lql_classifier_u32_helper_add_keys(u32, sel, U32_BUFF);
	}

	return(sel);
}

gboolean
lql_classifier_u32_helper_fill_from_nlhdr(LQLClassifierU32 *u32,
					  const struct nlmsghdr *hdr)
{
	struct tc_u32_sel *sel = NULL;
	__u32 *class = NULL;
	__u32 *hash = NULL;
	__u32 *link = NULL;
	__u32 *divisor = NULL;

	g_assert(u32 != NULL);
	g_assert(hdr != NULL);

	lql_classifier_set_ifIndex(LQL_CLASSIFIER(u32), nl_tcmsg_get_ifindex(hdr));
	lql_classifier_set_parentHandle(LQL_CLASSIFIER(u32), nl_tcmsg_get_parent(hdr));
	lql_classifier_set_id(LQL_CLASSIFIER(u32), nl_tcmsg_get_handle(hdr));
	lql_classifier_set_priority(LQL_CLASSIFIER(u32), nl_tcmsg_get_info(hdr) >> 16);
	lql_classifier_set_protocol(LQL_CLASSIFIER(u32),
				    (nl_tcmsg_get_info(hdr) & 0x0000FFFF));

	class = (unsigned int *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_U32_CLASSID);
	if (class != NULL) {
		lql_classifier_set_classHandle(LQL_CLASSIFIER(u32), *class);
	}

	hash = (unsigned int *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_U32_HASH);
	if (hash != NULL) {
		lql_classifier_u32_set_hash(LQL_CLASSIFIER_U32(u32), *hash);
	}

	link = (unsigned int *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_U32_LINK);
	if (link != NULL) {
		lql_classifier_u32_set_link(LQL_CLASSIFIER_U32(u32), *link);
	}

	divisor = (unsigned int *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_U32_DIVISOR);
	if (divisor != NULL) {
		lql_classifier_u32_set_divisor(LQL_CLASSIFIER_U32(u32), *divisor);
	}
		
	sel = (struct tc_u32_sel *)nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_U32_SEL);
	if (sel == NULL) {
//		g_printerr("No u32 sel poniter.\n");
	} else {
/* 		g_print("flags:%i\n", sel->flags); */
/* 		g_print("offshift:%i\n", sel->offshift); */
/* 		g_print("nkeys:%i\n", sel->nkeys); */
/* 		g_print("offmask:%i\n", sel->offmask); */
/* 		g_print("off:%i\n", sel->off); */
/* 		g_print("offoff:%i\n", sel->offoff); */
/* 		g_print("hoff:%i\n", sel->hoff); */
/* 		g_print("hmask:%i\n", sel->hmask); */
	}

	return(TRUE);
}

static gboolean
lql_classifier_u32_helper_do(LQLClassifierU32 *u32,
			     LQLCon *con,
			     int flags)
{
	U32Sel *sel = NULL;
	TcPacket *pkt = NULL;
	struct nlmsghdr *nlmsg = NULL;
	int len = 0;
	char *kind = "u32";
	guint32 class = 0, hash = 0, link = 0, divisor = 0;

	g_assert(con != NULL);
	g_assert(u32 != NULL);

	pkt = nl_tcpacket_new();
	nlmsg = nl_tcpacket_get_nlmsg(pkt);

	nl_nlmsg_set_type(nlmsg, RTM_NEWTFILTER);
	nl_nlmsg_set_flags(nlmsg, NLM_F_REQUEST | NLM_F_ACK | flags);
	nl_nlmsg_set_pid(nlmsg, getpid());
	nl_nlmsg_set_seq(nlmsg, lql_con_get_inc_sequence(LQL_CON(con)));

	nl_tcmsg_set_family(nlmsg, AF_UNSPEC);
	nl_tcmsg_set_ifindex(nlmsg, lql_classifier_get_ifIndex(LQL_CLASSIFIER(u32)));
	nl_tcmsg_set_parent(nlmsg, lql_classifier_get_parentHandle(LQL_CLASSIFIER(u32)));
	nl_tcmsg_set_handle(nlmsg, lql_classifier_get_id(LQL_CLASSIFIER(u32)));
	nl_tcmsg_set_info(nlmsg,
			  TC_H_MAKE(lql_classifier_get_priority(LQL_CLASSIFIER(u32))<<16,
				    lql_classifier_get_protocol(LQL_CLASSIFIER(u32))));

	nl_tcpacket_add_rtattr(pkt, TCA_KIND, strlen(kind) + 1, kind);

	/* Get the values that may need to be added as rtattrs. */
	if ((lql_classifier_u32_get_match_len(LQL_CLASSIFIER_U32(u32)) > 0) ||
	    lql_classifier_u32_get_hashMask(u32) ||
	    lql_classifier_u32_get_hashOffset(u32)) {
		sel = lql_classifier_u32_helper_build_sel(u32, &len);
	}
	class = lql_classifier_get_classHandle(LQL_CLASSIFIER(u32));
	link = lql_classifier_u32_get_link(LQL_CLASSIFIER_U32(u32));
	divisor = lql_classifier_u32_get_divisor(LQL_CLASSIFIER_U32(u32));

	if ((sel != NULL) || (class > 0) || (hash > 0) || (link > 0) || (divisor > 0)) {
		nl_tcpacket_add_rtattr(pkt, TCA_OPTIONS, 0, NULL);
	}

	if (sel != NULL) {
		nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_U32_SEL,
					   NLMSG_ALIGN(len), sel);
	}
	if (class > 0) {
		g_print("adding class\n");
		nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_U32_CLASSID,
					   sizeof(class), &class);
	}
	if (lql_classifier_get_id(LQL_CLASSIFIER(u32)) & 0x000FF000) {
		g_print("adding hash\n");
		hash = lql_classifier_get_id(LQL_CLASSIFIER(u32)) & 0xFFFFF000;
		nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_U32_HASH,
					   sizeof(hash), &hash);
	}
	if (link > 0) {
		g_print("adding link\n");
		nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_U32_LINK,
					   sizeof(link), &link);
	}
	if (divisor > 0) {
		g_print("adding divisor\n");
		nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_U32_DIVISOR,
					   sizeof(divisor), &divisor);
	}

	if (!nl_tcpacket_do_command(pkt, lql_con_get_nlcon(LQL_CON(con)))) {
		return(FALSE);
	}

	g_free(pkt);
	g_free(sel);
	
	return(TRUE);
}

gboolean
lql_classifier_u32_helper_add(LQLClassifierU32 *u32,
			      LQLCon *con)
{
	g_assert(u32 != NULL);
	g_assert(con != NULL);

	return(lql_classifier_u32_helper_do(u32, con, NLM_F_CREATE | NLM_F_EXCL));
}

gboolean
lql_classifier_u32_helper_modify(LQLClassifierU32 *u32,
				 LQLCon *con)
{
	g_assert(u32 != NULL);
	g_assert(con != NULL);

	return(lql_classifier_u32_helper_do(u32, con, 0));
}

gboolean
lql_classifier_u32_helper_add_match(LQLClassifierU32 *u32,
				    LQLU32Match *match)
{
	g_assert(u32 != NULL);
	g_assert(match != NULL);

	// Add a ref for while it's in our list.
	g_object_ref(G_OBJECT(match));

	u32->private->matchList = g_list_append(u32->private->matchList, match);

	return(TRUE);
}

gboolean
lql_classifier_u32_helper_add_match_ip_dst(LQLClassifierU32 *u32,
					   const gchar *ip)
{
	LQLU32Match *match = NULL;
	struct in_addr dst;

	g_assert(u32 != NULL);
	g_assert(ip != NULL);

	if (inet_pton(AF_INET, ip, &dst) <= 0) {
		// Conversion failed.
		return(FALSE);
	}

	match = (LQLU32Match *) lql_u32match_new();
	lql_u32match_set_value(match, dst.s_addr);
	lql_u32match_set_mask(match, 0xFFFFFFFF);
	lql_u32match_set_offset(match, IP_DST_OFFSET);

	if (!lql_classifier_u32_add_match(u32, match)) {
		return(FALSE);
	}

	g_object_unref(match);
	
	return(TRUE);
}

LQLU32Match*
lql_classifier_u32_helper_get_match(LQLClassifierU32 *u32,
				    int position)
{
	LQLU32Match *match = NULL;

	g_assert(u32 != NULL);
	g_assert(position >= 0);
	
	match = g_list_nth_data(u32->private->matchList, position);

	return(match);
}

int
lql_classifier_u32_helper_get_match_len(LQLClassifierU32 *u32)
{
	g_assert(u32 != NULL);
	
	return(g_list_length(u32->private->matchList));
}
