/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>

#include <asm/types.h>
#include <asm/param.h>
#include <sys/socket.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>

#include <glib.h>
#include <glib-object.h>

#include "debug_macros.h"
#include "nl.h"
#include "lql-con.h"
#include "lql-con-private.h"

#include "lql-interface.h"

#include "lql-qdisc.h"
#include "lql-qdisc-htb.h"
#include "lql-qdisc-pfifofast.h"
#include "lql-qdisc-pfifo.h"
#include "lql-qdisc-sfq.h"
#include "lql-qdisc-priomap.h"
#include "lql-qdisc-dsmark.h"
#include "lql-qdisc-netem.h"

#include "lql-class.h"
#include "lql-class-htb.h"
#include "lql-class-priomap.h"
#include "lql-class-dsmark.h"
#include "lql-class-netem.h"

#include "lql-classifier.h"
#include "lql-classifier-u32.h"
#include "lql-classifier-fwmark.h"
#include "lql-classifier-tcindex.h"

#include "lql-u32match.h"

#define BUFF 4096

/* Prototypes. */
int lql_con_helper_get_cell_log(int size);
gboolean lql_con_helper_calc_rate_table(LQLCon *con, unsigned bps, __u32 *rtab, unsigned mtu, unsigned mpu);
unsigned lql_con_helper_calc_xmit_time(LQLCon *con, unsigned bps, unsigned size);
LQLInterface* lql_con_helper_get_interface(LQLCon *con, int index);
static GList * lql_con_helper_get_type(LQLCon *con, unsigned int reqType, int ifIndex);

/* Brutal global hack. */
double tick = 0;

gboolean
lql_con_helper_init(LQLCon *con)
{
	FILE *fp = NULL;
	g_assert(con != NULL);
	__u32 tpus, uspt;

	con->private->con = nl_open(0);
	if (con->private->con == NULL) {
		g_printerr("Could not open nl connection.\n");
		return(FALSE);
	}

	fp = fopen("/proc/net/psched", "r");
	if (fp == NULL) {
		g_printerr("Could not open /proc/net/psched\n");
		return(FALSE);
	}

	if (fscanf(fp, "%08x%08x", &tpus, &uspt) != 2) {
		fclose(fp);
		return(FALSE);
	}

	con->private->tickInUsec = (double)tpus/uspt;
	//lql_con_set_tickInUsec(con, (double)tpus/uspt);
	con->private->hz = HZ;
	//lql_con_set_hz(con, HZ);

	/* Brual global hack for testing. */
	tick = con->private->tickInUsec;

	fclose(fp);

	return(TRUE);
}

int
lql_con_helper_get_cell_log(int size)
{
	int cell_log = 0;

	while ((size >> cell_log) > 255) {
		cell_log++;
	}

	return(cell_log);
}

static long
lql_con_helper_calc_usec2tick(LQLCon *con,
			      long usec)
{
	g_assert(con != NULL);

	return(lql_con_get_tickInUsec(con) * usec);
}

unsigned
lql_con_helper_calc_xmit_time(LQLCon *con,
			      unsigned bps,
			      unsigned size)
{
	g_assert(con != NULL);
	g_assert(bps > 0);
	g_assert(size > 0);
	
	return(lql_con_helper_calc_usec2tick(con, 1000000 * ((double)size/bps)));
}

gboolean
lql_con_helper_calc_rate_table(LQLCon *con,
			       unsigned bps,
			       __u32 *rtab,
			       unsigned mtu,
			       unsigned mpu)
{
	int i, cell_log;
       
	g_assert(con != NULL);
	g_assert(bps > 0);
	g_assert(rtab != NULL);
	g_assert(mtu > 0);
	g_assert(mpu > 0);

	cell_log = lql_con_helper_get_cell_log(mtu);
	
	for (i = 0; i < 256; i++) {
		unsigned size = (i << cell_log);
		if (size < mpu) {
			size = mpu;
		}
		
		rtab[i] = lql_con_helper_calc_xmit_time(con, bps, size);
	}

	return(TRUE);
}

static LQLInterface*
lql_con_helper_build_interface(const struct nlmsghdr *hdr)
{
	LQLInterface *interface = NULL;

	g_assert(hdr != NULL);

	interface = (LQLInterface*)lql_interface_new();
	if(!lql_interface_fill_from_nlhdr(interface, hdr)) {
		return(NULL);
	} else {
		return(interface);
	}
}

static LQLQDisc*
lql_con_helper_build_qdisc(const struct nlmsghdr *hdr)
{
	gchar *type = NULL;
	LQLQDisc *qdisc = NULL;

	g_assert(hdr != NULL);

	type = (char *)nl_tcmsg_get_rta_val(hdr, TCA_KIND);

	if (g_strcasecmp("htb", type) == 0) {
		qdisc = (LQLQDisc *)lql_qdisc_htb_new();
		if (!lql_qdisc_fill_from_nlhdr(LQL_QDISC(qdisc), hdr)) {
			// Need to obj_unref each of these cases.
			return(NULL);
		} else {
			return(qdisc);
		}
	} else if (g_strcasecmp("sfq", type) == 0) {
		qdisc = (LQLQDisc *)lql_qdisc_sfq_new();
		if (!lql_qdisc_fill_from_nlhdr(LQL_QDISC(qdisc), hdr)) {
			return(NULL);
		} else {
			return(qdisc);
		}
	} else if (g_strcasecmp("pfifo_fast", type) == 0) {
		qdisc = (LQLQDisc *)lql_qdisc_pfifofast_new();
		if (!lql_qdisc_fill_from_nlhdr(LQL_QDISC(qdisc), hdr)) {
			return(NULL);
		} else {
			return(qdisc);
		}
	} else if (g_strcasecmp("pfifo", type) == 0) {
		qdisc = (LQLQDisc *)lql_qdisc_pfifo_new();
		if (!lql_qdisc_fill_from_nlhdr(LQL_QDISC(qdisc), hdr)) {
			return(NULL);
		} else {
			return(qdisc);
		}
	} else if (g_strcasecmp("prio", type) == 0) {
		qdisc = (LQLQDisc *)lql_qdisc_priomap_new();
		if (!lql_qdisc_fill_from_nlhdr(LQL_QDISC(qdisc), hdr)) {
			return(NULL);
		} else {
			return(qdisc);
		}
	} else if (g_strcasecmp("dsmark", type) == 0) {
		qdisc = (LQLQDisc *)lql_qdisc_dsmark_new();
		if (!lql_qdisc_fill_from_nlhdr(LQL_QDISC(qdisc), hdr)) {
			return(NULL);
		} else {
			return(qdisc);
		}
	} else if (g_strcasecmp("netem", type) == 0) {
		qdisc = (LQLQDisc *)lql_qdisc_netem_new();
		if (!lql_qdisc_fill_from_nlhdr(LQL_QDISC(qdisc), hdr)) {
			return(NULL);
		} else {
			return(qdisc);
		}
	} else {
		g_printerr("Unknown qdisc type.\n");
	}
	
	return(NULL);
}

static LQLClass*
lql_con_helper_build_class(const struct nlmsghdr *hdr)
{
	gchar *type = NULL;
	LQLClass *class = NULL;
	
	g_assert(hdr != NULL);

	type = (char *) nl_tcmsg_get_rta_val(hdr, TCA_KIND);
	
	if (g_strcasecmp("htb", type) == 0) {
		class = LQL_CLASS(lql_class_htb_new());
		if (!lql_class_fill_from_nlhdr(LQL_CLASS(class), hdr)) {
			return(NULL);
		} else {
			return(class);
		}
	} else if (g_strcasecmp("prio", type) == 0) {
		class = LQL_CLASS(lql_class_priomap_new());
		if (!lql_class_fill_from_nlhdr(LQL_CLASS(class), hdr)) {
			return(NULL);
		} else {
			return(class);
		}
	} else if (g_strcasecmp("dsmark", type) == 0) {
		class = LQL_CLASS(lql_class_dsmark_new());
		if (!lql_class_fill_from_nlhdr(LQL_CLASS(class), hdr)) {
			return(NULL);
		} else {
			return(class);
		}
	} else if (g_strcasecmp("netem", type) == 0) {
		class = LQL_CLASS(lql_class_netem_new());
		if (!lql_class_fill_from_nlhdr(LQL_CLASS(class), hdr)) {
			return(NULL);
		} else {
			return(class);
		}
	} else {
		g_printerr("Unknown class type.\n");
	}
	
	return(NULL);
}

static LQLClassifier*
lql_con_helper_build_classifier(const struct nlmsghdr *hdr)
{
	gchar *type = NULL;
	LQLClassifier *classifier = NULL;
	
	g_assert(hdr != NULL);
	
	type = (char *)nl_tcmsg_get_rta_val(hdr, TCA_KIND);

	if (g_strcasecmp("u32", type) == 0) {
		classifier = (LQLClassifier *)lql_classifier_u32_new();
		if (!lql_classifier_fill_from_nlhdr(LQL_CLASSIFIER(classifier), hdr)) {
			return(NULL);
		} else {
			return(classifier);
		}
	} else if (g_strcasecmp("fw", type) == 0) {
		classifier = (LQLClassifier *)lql_classifier_fwmark_new();
		if (!lql_classifier_fill_from_nlhdr(LQL_CLASSIFIER(classifier), hdr)) {
			return(NULL);
		} else {
			return(classifier);
		}
	} else if (g_strcasecmp("tcindex", type) == 0) {
		classifier = (LQLClassifier *)lql_classifier_tcindex_new();
		if (!lql_classifier_fill_from_nlhdr(LQL_CLASSIFIER(classifier), hdr)) {
			return(NULL);
		} else {
			return(classifier);
		}
	} else {
		g_printerr("Unknown classifier type.\n");
	}
	
	return(NULL);
}

GList *
lql_con_helper_get_interfaces(LQLCon *con)
{
	IfInfoPacket *pkt = NULL;
	struct nlmsghdr *hdr = NULL;
	const struct nlmsghdr *hdrConst = NULL;
	char *buf = NULL;
	int i = 0, ret = 0;
	GList *list = NULL;

	g_assert(con != NULL);

	pkt = nl_ifinfopacket_new();
	hdr = nl_ifinfopacket_get_nlmsg(pkt);
	
	nl_nlmsg_set_type(hdr, RTM_GETLINK);
	nl_nlmsg_set_flags(hdr, NLM_F_REQUEST | NLM_F_ACK | NLM_F_MATCH);
	nl_nlmsg_set_pid(hdr, getpid());
	nl_nlmsg_set_seq(hdr, lql_con_get_inc_sequence(con));

	if (!nl_send(con->private->con, pkt, nl_nlmsg_get_length(hdr))) {
		return(NULL);
	}
	
	buf = g_malloc0(BUFF);

	for (;;) {
		ret = nl_read(con->private->con, buf, BUFF);
		i = 0;

		while ((hdrConst = nl_get_nlmsghdr(buf, ret, i)) != NULL) {
			if (hdrConst->nlmsg_type == RTM_NEWLINK) {
				list = g_list_append(list,
						     lql_con_helper_build_interface(hdrConst));
			}

			if (nl_nlmsg_is_done(hdrConst)) {
				goto done;
			}

			i++;
		}
	}

 done:
	g_free(buf);
	g_free(pkt);

	return(list);
}

GList *
lql_con_helper_get_qdiscs(LQLCon *con,
			  int index)
{
	GList *list = NULL;

	g_assert(con != NULL);
	g_assert(index > 0);

	list = lql_con_helper_get_type(con, RTM_GETQDISC, index);
		
	return(list);
}

GList *
lql_con_helper_get_classes(LQLCon *con,
			   int index)
{
	GList *list = NULL;

	g_assert(con != NULL);
	g_assert(index > 0);

	list = lql_con_helper_get_type(con, RTM_GETTCLASS, index);
	
	return(list);
}

GList*
lql_con_helper_get_classifiers(LQLCon *con,
			       int index)
{
	GList *list = NULL;

	g_assert(con != NULL);
	g_assert(index > 0);

	list = lql_con_helper_get_type(con, RTM_GETTFILTER, index);
	
	return(list);
}

static GList *
lql_con_helper_get_type(LQLCon *con,
			unsigned int reqType,
			int ifIndex)
{
	TcPacket *pkt = NULL;
	struct nlmsghdr *hdr = NULL;
	const struct nlmsghdr *hdrConst = NULL;
	char *buf = NULL;
	int i = 0, ret = 0;
	GList *list = NULL;

	g_assert(con != NULL);

	pkt = nl_tcpacket_new();
	hdr = nl_tcpacket_get_nlmsg(pkt);
	
	nl_nlmsg_set_type(hdr, reqType);
	nl_nlmsg_set_flags(hdr, NLM_F_REQUEST | NLM_F_ACK | NLM_F_MATCH);
	nl_nlmsg_set_pid(hdr, getpid());
	nl_nlmsg_set_seq(hdr, lql_con_get_inc_sequence(con));

	nl_tcmsg_set_ifindex(hdr, ifIndex);

	if (!nl_send(con->private->con, pkt, nl_nlmsg_get_length(hdr))) {
		return(NULL);
	}
	
	buf = g_malloc0(BUFF);

	for (;;) {
		ret = nl_read(con->private->con, buf, BUFF);
		i = 0;

		while ((hdrConst = nl_get_nlmsghdr(buf, ret, i)) != NULL) {
			/* Filter for the interface we want. */
			if (nl_tcmsg_get_ifindex(hdrConst) == ifIndex) {
			}

			if (hdrConst->nlmsg_type == RTM_NEWQDISC) {
				if (nl_tcmsg_get_ifindex(hdrConst) == ifIndex) {
					list = g_list_append(list,
							     lql_con_helper_build_qdisc(hdrConst));
				}
			} else if (hdrConst->nlmsg_type == RTM_NEWTCLASS) {
				if (nl_tcmsg_get_ifindex(hdrConst) == ifIndex) {
					list = g_list_append(list,
							     lql_con_helper_build_class(hdrConst));
				}
			} else if (hdrConst->nlmsg_type == RTM_NEWTFILTER) {
				if (nl_tcmsg_get_ifindex(hdrConst) == ifIndex) {
					list = g_list_append(list,
							     lql_con_helper_build_classifier(hdrConst));
				}
			} else if (hdrConst->nlmsg_type == NLMSG_DONE) {
				// Just so DONE messages don't cause the below print.
			} else {
				g_printerr("Unexpected type:%i\n", hdrConst->nlmsg_type);
			}

			if (nl_nlmsg_is_done(hdrConst)) {
				goto done;
			}

			i++;
		}
	}

 done:
	g_free(buf);
	g_free(pkt);

	return(list);
}

LQLInterface*
lql_con_helper_get_interface(LQLCon *con,
			     int index)
{
	GList *list = NULL;
	LQLInterface *interface = NULL, *ret = NULL;

	g_assert(con != NULL);
	g_assert(index > 0);

	list = lql_con_helper_get_interfaces(con);

	while (list != NULL) {
		interface = (LQLInterface *)list->data;
		if (lql_interface_get_ifIndex(LQL_INTERFACE(interface)) == index) {
			// The one we need.
			ret = interface;
		} else {
			// Don't need it.
			g_object_unref(G_OBJECT(interface));
		}

		list = list->next;
	}

	return(ret);
}
