/*
 * Copyright (C) 2004 Dan Siemon <dan@coverfire.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <unistd.h>
#include <string.h>

#include "debug_macros.h"
#include "nl.h"
#include "lql-con.h"
#include "lql-qdisc-dsmark.h"
#include "lql-qdisc-dsmark-private.h"

gboolean
lql_qdisc_dsmark_helper_fill_from_nlhdr(LQLQDiscDSMark *qdisc,
					const struct nlmsghdr *hdr)
{
	guint16 *tmp = NULL;

	g_assert(qdisc != NULL);
	g_assert(hdr != NULL);

	lql_element_fill_from_nlhdr(LQL_ELEMENT(qdisc), hdr);

	tmp = (guint16 *) nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_DSMARK_INDICES);
	if (tmp != NULL) {
		lql_qdisc_dsmark_set_indices(qdisc, *tmp);
	}

	tmp = (guint16 *) nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_DSMARK_DEFAULT_INDEX);
	if (tmp != NULL) {
		lql_qdisc_dsmark_set_defaultIndex(qdisc, *tmp);
	}

	if (nl_tcmsg_get_rta_sub_val(hdr, TCA_OPTIONS, TCA_DSMARK_SET_TC_INDEX) == NULL) {
		lql_qdisc_dsmark_set_setTCIndex(qdisc, FALSE);
	} else {
		lql_qdisc_dsmark_set_setTCIndex(qdisc, TRUE);
	}
		
	return(TRUE);
}

static gboolean
lql_qdisc_dsmark_helper_do(LQLQDiscDSMark *qdisc,
			   LQLCon *con,
			   int flags)
{
	TcPacket *pkt = NULL;
	struct nlmsghdr *hdr = NULL;
	gchar *kind = "dsmark";
	guint16 indices = 0, defaultIndex = 0;

	g_assert(con != NULL);
	g_assert(qdisc != NULL);

	pkt = nl_tcpacket_new();
	hdr = nl_tcpacket_get_nlmsg(pkt);
	
	nl_nlmsg_set_type(hdr, RTM_NEWQDISC);
	nl_nlmsg_set_flags(hdr, NLM_F_REQUEST | NLM_F_ACK | flags);
	nl_nlmsg_set_pid(hdr, getpid());
	nl_nlmsg_set_seq(hdr, lql_con_get_inc_sequence(LQL_CON(con)));

	nl_tcmsg_set_family(hdr, AF_UNSPEC);
	nl_tcmsg_set_ifindex(hdr, lql_element_get_ifIndex(LQL_ELEMENT(qdisc)));
	nl_tcmsg_set_parent(hdr, lql_element_get_parentHandle(LQL_ELEMENT(qdisc)));
	nl_tcmsg_set_handle(hdr, lql_element_get_handle(LQL_ELEMENT(qdisc)));
	
	if (!nl_tcpacket_add_rtattr(pkt, TCA_KIND, strlen(kind) + 1, kind)) {
		debug_print(5, "Could not add TCA_KIND");
	}

	if (!nl_tcpacket_add_rtattr(pkt, TCA_OPTIONS, 0, NULL)) {
		debug_print(5, "Could not add TCA_OPTIONS");
	}

	indices = lql_qdisc_dsmark_get_indices(qdisc);
	if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_DSMARK_INDICES,
					NLMSG_ALIGN(sizeof(__u16)), &indices)) {
		debug_print(5, "Could not add TCA_DSMARK_INDICES\n");
	}

	defaultIndex = lql_qdisc_dsmark_get_defaultIndex(qdisc);
	if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_DSMARK_DEFAULT_INDEX,
					NLMSG_ALIGN(sizeof(__u16)), &defaultIndex)) {
		debug_print(5, "Could not add TCA_DSMARK_DEFAULT_INDEX\n");
	}

	if (lql_qdisc_dsmark_get_setTCIndex(qdisc)) {
		/* It's true so add the options. */
		if (!nl_tcpacket_add_rtattr_sub(pkt, TCA_OPTIONS, TCA_DSMARK_SET_TC_INDEX,
						0, NULL)) {
			debug_print(5, "Could not add TCA_DSMARK_DEFAULT_INDEX\n");
		}
	}
	
	if (!nl_tcpacket_do_command(pkt, lql_con_get_nlcon(LQL_CON(con)))) {
		return(FALSE);
	}
	
	g_free(pkt);
	
	return(TRUE);
}

gboolean
lql_qdisc_dsmark_helper_add(LQLQDiscDSMark *qdisc,
			    LQLCon *con)
{
	return(lql_qdisc_dsmark_helper_do(qdisc, con, NLM_F_CREATE | NLM_F_EXCL));
}

gboolean
lql_qdisc_dsmark_helper_modify(LQLQDiscDSMark *qdisc,
			       LQLCon *con)
{
	return(FALSE);
}

void
lql_qdisc_dsmark_helper_print(LQLQDisc *self)
{
	LQLQDiscDSMarkClass *klass = NULL;
	LQLQDiscClass *parentKlass = NULL;
	LQLQDiscDSMark *dsmark = NULL;

	g_assert(self != NULL);
	g_assert(LQL_IS_QDISC_DSMARK(self));

	dsmark = LQL_QDISC_DSMARK(self);

        if (dsmark->private->dispose_has_run) {
                return;
        }

	klass = LQL_QDISC_DSMARK_GET_CLASS(self);
	parentKlass = g_type_class_peek_parent(klass);
	
	parentKlass->print(self);

	g_print("Type: DSMark\n");
	g_print("DSMark indices: %u\n", lql_qdisc_dsmark_get_indices(dsmark));
	g_print("DSMark default index: %u\n", lql_qdisc_dsmark_get_defaultIndex(dsmark));
	if (lql_qdisc_dsmark_get_setTCIndex(dsmark)) {
		g_print("DSMark set_tc_index: TRUE\n");
	} else {
		g_print("DSMark set_tc_index: FALSE\n");
	}
}
