#include <glib.h>

#include "lql-con.h"
#include "lql-interface.h"
#include "lql-qdisc-htb.h"
#include "lql-class-htb.h"
#include "lql-classifier-fwmark.h"

int
main(int argc,
     char *argv[])
{
	LQLCon *con = NULL;
	LQLQDiscHTB *qdisc = NULL;
	LQLClassHTB *class1 = NULL;
	GList *list = NULL;
	LQLInterface *interface = NULL;
	LQLClassifierFwmark *fwmark = NULL;

	g_type_init();

	con = lql_con_new();

	interface = lql_con_get_interface_by_name(con, "eth0");
	if (interface == NULL) {
		g_printerr("Could not get interface.\n");
		return(1);
	}

	/******************* Add ******************/

	g_print("Adding a root qdisc.\n");
	
	qdisc = lql_qdisc_htb_new();
	lql_qdisc_set_parent(LQL_QDISC(qdisc), NULL);
	lql_qdisc_set_id(LQL_QDISC(qdisc), 1);
	lql_qdisc_htb_set_defaultClass(LQL_QDISC_HTB(qdisc), 1);
	lql_qdisc_htb_set_rateToQuantum(LQL_QDISC_HTB(qdisc), 20);
	lql_element_set_interface(LQL_ELEMENT(qdisc), interface);

	if (!lql_qdisc_add(LQL_QDISC(qdisc), con)) {
		g_printerr("Could not add qdisc.\n");
		return(1);
	}

	g_print("Adding class 1.\n");

	class1 = lql_class_htb_new();
	lql_class_set_parent(LQL_CLASS(class1), LQL_ELEMENT(qdisc));
	lql_class_set_priority(LQL_CLASS(class1), 6);
	lql_class_set_id(LQL_CLASS(class1), 1);
	lql_element_set_interface(LQL_ELEMENT(class1), interface);
	lql_class_htb_set_rate(class1, 1000);
	lql_class_htb_set_ceiling(class1, 1001);
	if (!lql_class_add(LQL_CLASS(class1), con)) {
		g_printerr("Line #%i could not add class.\n", __LINE__);
		return(1);
	}
	g_object_unref(class1);

	/***************** Verify QDisc ****************/

	g_print("Verifying successful qdisc add.\n");

	list = lql_con_get_qdiscs(con, interface);
	if (list == NULL) {
		g_printerr("Could not get qdisc list.\n");
		return(1);
	}

	if (g_list_length(list) != 1) {
		g_printerr("Wrong number of qdiscs returned.\n");
		return(1);
	}

	qdisc = LQL_QDISC_HTB(list->data);
	
	if (lql_element_get_handle(LQL_ELEMENT(qdisc)) != 65536) {
		g_printerr("Wrong ID number. Got: %u\n",
			   lql_element_get_handle(LQL_ELEMENT(qdisc)));
		return(1);
	}

	if (lql_element_get_ifIndex(LQL_ELEMENT(qdisc)) !=
	    lql_interface_get_ifIndex(interface)) {
		g_printerr("Wrong interface ID.\n");
		return(1);
	}

	if (lql_element_get_parentHandle(LQL_ELEMENT(qdisc)) != TC_H_ROOT) {
		g_printerr("Line #%i wrong parent handle.\n", __LINE__);
		return(1);
	}

	if (lql_qdisc_htb_get_defaultClass(qdisc) != 1) {
		g_printerr("Wrong default class.\n");
		return(1);
	}
	
	if (lql_qdisc_htb_get_rateToQuantum(qdisc) != 20) {
		g_printerr("Wrong rate to quantum.\n");
		return(1);
	}

	g_list_free(list); list = NULL;

	/*************** Verify classes *****************/

	g_print("Verifying classes.\n");

	list = lql_con_get_classes(con, interface);
	if (list == NULL) {
		g_printerr("Line #%i could not get class list.\n", __LINE__);
		return(1);
	}
	
	if (g_list_length(list) != 1) {
		g_printerr("Line #%i wrong number of classes returned.\n", __LINE__);
		return(1);
	}

	/*********** class 1 ************/

	g_print("Verifying class 1.\n");
	
	class1 = LQL_CLASS_HTB(list->data);

	if (lql_class_get_id(LQL_CLASS(class1)) != 1) {
		g_printerr("Line #%i wrong id.\n", __LINE__);
		return(1);
	}
	
	if (lql_class_get_priority(LQL_CLASS(class1)) != 6) {
		g_printerr("Line #%i wrong priority. Got %u\n",
			   __LINE__, lql_class_get_priority(LQL_CLASS(class1)));
		return(1);
	}
	
	if (lql_element_get_ifIndex(LQL_ELEMENT(class1)) !=
	    lql_interface_get_ifIndex(interface)) {
		g_printerr("Line #%i wrong interface index.\n", __LINE__);
		return(1);
	}
	
	if (lql_class_htb_get_rate(class1) != 1000) {
		g_printerr("Line #%i wrong rate.\n", __LINE__);
		return(1);
	}
	
	if (lql_class_htb_get_ceiling(class1) != 1001) {
		g_printerr("Line #%i wrong ceil.\n", __LINE__);
		return(1);
	}

	/*********** Adding fwmark classifier ***************/
	fwmark = (LQLClassifierFwmark *)lql_classifier_fwmark_new();
	lql_classifier_set_priority(LQL_CLASSIFIER(fwmark), 1);
	lql_classifier_set_interface(LQL_CLASSIFIER(fwmark), LQL_INTERFACE(interface));
	lql_classifier_set_class(LQL_CLASSIFIER(fwmark), LQL_CLASS(class1));
	lql_classifier_set_parent(LQL_CLASSIFIER(fwmark), LQL_ELEMENT(qdisc));
	lql_classifier_fwmark_set_mark(LQL_CLASSIFIER_FWMARK(fwmark), 5);
	if (!lql_classifier_add(LQL_CLASSIFIER(fwmark), con)) {
		g_printerr("Line #%i could not add classifier.\n", __LINE__);
		return(-1);
	}
	g_object_unref(fwmark);

	/************ Verify classifier *********************/
	list = lql_con_get_classifiers(con, interface);
	
	if (g_list_length(list) != 2) {
		g_printerr("Line #%i wrong number of classifiers. Got %i\n",
			   __LINE__, g_list_length(list));
		return(1);
	}
	
	fwmark = LQL_CLASSIFIER_FWMARK(list->next->data);
	
	if (lql_classifier_get_priority(LQL_CLASSIFIER(fwmark)) != 1) {
		g_printerr("Line #%i wrong priority value.\n", __LINE__);
		return(1);
	}
	
	if (lql_classifier_get_ifIndex(LQL_CLASSIFIER(fwmark)) !=
	    lql_interface_get_ifIndex(interface)) {
		g_printerr("Line #%i wrong interface value.\n", __LINE__);
		return(1);
	}

	if (lql_classifier_get_classHandle(LQL_CLASSIFIER(fwmark)) !=
	    lql_element_get_handle(LQL_ELEMENT(class1))) {
		g_printerr("Line #%i wrong class value.\n", __LINE__);
		return(1);
	}
	
	if (lql_classifier_get_parentHandle(LQL_CLASSIFIER(fwmark)) !=
	    lql_element_get_handle(LQL_ELEMENT(qdisc))) {
		g_printerr("Line #%i wrong parent handle.\n", __LINE__);
		return(1);
	}
	
	if (lql_classifier_fwmark_get_mark(fwmark) != 5) {
		g_printerr("Line #%i wrong mark value.\n", __LINE__);
		return(1);
	}

	/********** Deleting classifier ***********/
	g_print("Deleting classifier.\n");
	if (!lql_classifier_delete(LQL_CLASSIFIER(fwmark), con)) {
		g_printerr("Line #%i could not delete classifier.\n", __LINE__);
		return(1);
	}

	/********** Verify deletion **************/
	g_print("Verifying deletion.\n");
	list = lql_con_get_classifiers(con, interface);
	
	if (g_list_length(list) != 0) {
		g_printerr("Line #%i wrong number of classifiers. Got %i\n",
			   __LINE__, g_list_length(list));
		return(1);
	}

        /*********** Delete class **************/
	g_print("Deleting class.\n");
	if (!lql_class_delete(LQL_CLASS(class1), con)) {
		g_printerr("Line #%i could not delete class.\n", __LINE__);
		return(1);
	}
	g_object_unref(class1);
	g_list_free(list);

	/************ Delete the QDisc. ****************/
	g_print("Deleting qdisc.\n");
	
	if (!lql_qdisc_delete(LQL_QDISC(qdisc), con)) {
		g_printerr("Line #%i could not delete qdisc.\n", __LINE__);
		return(1);
	}

	return(0);
}
