#include <glib.h>

#include "lql-con.h"
#include "lql-interface.h"
#include "lql-qdisc-dsmark.h"
#include "lql-classifier-tcindex.h"

int
main(int argc,
     char *argv[])
{
	LQLCon *con = NULL;
	LQLQDiscDSMark *qdisc = NULL;
	LQLClassifierTCIndex *tcindex = NULL;
	GList *list = NULL;
	LQLInterface *interface = NULL;

	g_type_init();

	con = lql_con_new();

	interface = lql_con_get_interface_by_name(con, "eth0");
	if (interface == NULL) {
		g_printerr("Could not get interface.\n");
		return(1);
	}

	g_print("Adding a root qdisc.\n");

	qdisc = lql_qdisc_dsmark_new();
	lql_qdisc_set_parent(LQL_QDISC(qdisc), NULL);
	lql_qdisc_set_id(LQL_QDISC(qdisc), 1);
	lql_element_set_interface(LQL_ELEMENT(qdisc), interface);
	lql_qdisc_dsmark_set_indices(qdisc, 32);
	lql_qdisc_dsmark_set_setTCIndex(qdisc, TRUE);
	lql_qdisc_dsmark_set_defaultIndex(qdisc, 2);

	if (!lql_qdisc_add(LQL_QDISC(qdisc), con)) {
		g_printerr("Line #%i could not add qdisc.\n", __LINE__);
		return(1);
	}

	g_object_unref(qdisc);

	/************** Verify *******************/

	g_print("Verifying successful add.\n");

	list = lql_con_get_qdiscs(con, interface);
	if (list == NULL) {
		g_printerr("Could not get qdisc list.\n");
		return(1);
	}

	if (g_list_length(list) != 1) {
		g_printerr("Wrong number of qdiscs returned.\n");
		return(1);
	}

	qdisc = LQL_QDISC_DSMARK(list->data);

	if (lql_element_get_handle(LQL_ELEMENT(qdisc)) != 65536) {
		g_printerr("Wrong ID number. Got: %u\n",
			   lql_element_get_handle(LQL_ELEMENT(qdisc)));
		return(1);
	}

	if (lql_element_get_ifIndex(LQL_ELEMENT(qdisc)) !=
	    lql_interface_get_ifIndex(interface)) {
		g_printerr("Wrong interface ID.\n");
		return(1);
	}

	if (lql_element_get_parentHandle(LQL_ELEMENT(qdisc)) != TC_H_ROOT) {
		g_printerr("Line #%i wrong parent handle.\n", __LINE__);
		return(1);
	}

	if (lql_qdisc_dsmark_get_indices(qdisc) != 32) {
		g_printerr("Line #%i returned indices value is wrong. Got %u\n",
			   __LINE__, lql_qdisc_dsmark_get_indices(qdisc));
		return(1);
	}
	
	if (lql_qdisc_dsmark_get_setTCIndex(qdisc) != TRUE) {
		g_printerr("Line #%i returned setTCIndex value is wrong.\n", __LINE__);
		return(1);
	}

	if (lql_qdisc_dsmark_get_defaultIndex(qdisc) != 2) {
		g_printerr("Line #%i returned default value is wrong.\n", __LINE__);
		return(1);
	}

	/************** Add classifier ***********************/
	g_print("Adding tcindex classifier.\n");

	tcindex = lql_classifier_tcindex_new();
	lql_classifier_set_id(LQL_CLASSIFIER(tcindex), 10);
	lql_classifier_set_interface(LQL_CLASSIFIER(tcindex), interface);
	lql_classifier_set_priority(LQL_CLASSIFIER(tcindex), 5);
	lql_classifier_set_parent(LQL_CLASSIFIER(tcindex), LQL_ELEMENT(qdisc));
	lql_classifier_tcindex_set_hash(tcindex, 0xff);
	lql_classifier_tcindex_set_mask(tcindex, 0xfc);
	lql_classifier_tcindex_set_shift(tcindex, 2);
	lql_classifier_tcindex_set_fallThrough(tcindex, TRUE);

	if (!lql_classifier_add(LQL_CLASSIFIER(tcindex), con)) {
		g_printerr("Line #%i could not add classifier.\n", __LINE__);
		return(1);
	}

	g_object_unref(tcindex);

	/********* Verify classifier ***********************/
	g_print("Verifying classifier.\n");

	list = lql_con_get_classifiers(LQL_CON(con), interface);
	
	if (g_list_length(list) != 1) {
		g_printerr("Line #%i wrong number of classifiers.\n", __LINE__);
		return(1);
	}

	tcindex = LQL_CLASSIFIER_TCINDEX(list->data);

	if (lql_classifier_get_ifIndex(LQL_CLASSIFIER(tcindex)) !=
	    lql_interface_get_ifIndex(interface)) {
		g_printerr("Line #%i wrong interface index.\n", __LINE__);
		return(1);
	}

	if (lql_classifier_get_priority(LQL_CLASSIFIER(tcindex)) != 5) {
		g_printerr("Line #%i wrong priority value.\n", __LINE__);
		return(1);
	}
	
	if (lql_classifier_get_parentHandle(LQL_CLASSIFIER(tcindex)) !=
	    lql_element_get_handle(LQL_ELEMENT(qdisc))) {
		g_printerr("Line #%i wrong parent value.\n", __LINE__);
		return(1);
	}
	
	if (lql_classifier_tcindex_get_hash(tcindex) != 0xff) {
		g_printerr("Line #%i wrong hash value. Got %u.\n",
			   __LINE__, lql_classifier_tcindex_get_hash(tcindex));
		return(1);
	}
	
	if (lql_classifier_tcindex_get_mask(tcindex) != 0xfc) {
		g_printerr("Line #%i wrong mask value.\n", __LINE__);
		return(1);
	}

	if (lql_classifier_tcindex_get_shift(tcindex) != 2) {
		g_printerr("Line #%i wrong shift value.\n", __LINE__);
		return(1);
	}

	if (lql_classifier_tcindex_get_fallThrough(tcindex) != TRUE) {
		g_printerr("Line #%i wrong fall through value.\n", __LINE__);
		return(1);
	}

	/************ Deleting classifier. ******************/
	g_print("Deleting classifier.\n");
	if (!lql_classifier_delete(LQL_CLASSIFIER(tcindex), con)) {
		g_printerr("Line #%i could not delete classifier.\n", __LINE__);
		return(1);
	}
	g_object_unref(tcindex);

	/************* Verify classifier is gone. **************/
	g_print("Verifying classifer is gone.\n");
	list = lql_con_get_classifiers(LQL_CON(con), interface);
	
	if (g_list_length(list) != 0) {
		g_printerr("Line #%i wrong number of classifiers.\n", __LINE__);
		return(1);
	}	

	/********** Delete QDisc. ***************/
	g_print("Deleting QDisc.\n");
	
	if (!lql_qdisc_delete(LQL_QDISC(qdisc), con)) {
		g_printerr("Line #%i could not delete qdisc.\n", __LINE__);
		return(1);
	}

	return(0);
}
